import math
import tkinter as tk
from tkinter import ttk

from thonny import get_workbench


class GridTable(tk.Frame):
    def __init__(self, master, header_rows, data_row_count, footer_row_count, frozen_column_count):
        super().__init__(master)

        self.header_widgets = {}
        self.data_widgets = {}

        self.bind("<Configure>", self.on_configure, True)

        self.screen_row_height = 22  # TODO:

        self.first_visible_data_row_no = 0
        self.visible_data_row_count = 0
        self.header_rows = header_rows
        self.data_rows = {}

        self.screen_row_count = 0
        self.data_row_count = data_row_count
        self.column_count = len(self.header_rows[-1])
        self.header_row_count = len(header_rows)
        self.footer_row_count = footer_row_count
        self.frozen_column_count = frozen_column_count

        self.update_header_rows()

    def set_data(self, data_rows):
        # self.data_rows.update(data_rows) # dict version
        self.data_rows = data_rows
        self.data_row_count = len(data_rows)
        self.update_screen_data()

    def update_header_rows(self):
        for row_no in range(self.header_row_count):
            for col_no in range(self.column_count):
                w = self.get_header_widget(self.screen_row_count, col_no)
                w.grid(row=row_no, column=col_no, sticky="nsew", pady=(0, 1), padx=(0, 1))
                w.configure(text=self.get_header_value(row_no, col_no))

        self.screen_row_count = self.header_row_count

    def get_data_widget(self, screen_row_no, col_no):
        if (screen_row_no, col_no) not in self.data_widgets:
            self.data_widgets[(screen_row_no, col_no)] = self.create_data_widget(col_no)

        return self.data_widgets[(screen_row_no, col_no)]

    def get_header_widget(self, row_no, col_no):
        if (row_no, col_no) not in self.header_widgets:
            self.header_widgets[(row_no, col_no)] = self.create_header_widget(col_no)

        return self.header_widgets[(row_no, col_no)]

    def create_data_widget(self, col_no):
        if col_no < self.frozen_column_count:
            background = None
        else:
            background = "white"

        return tk.Label(self, background=background, anchor="e", padx=7, text="")

    def create_header_widget(self, col_no):
        return tk.Label(self, anchor="e", padx=7, text="")

    def set_first_visible_data_row_no(self, n):
        self.first_visible_data_row_no = max(min(n, self.data_row_count), 0)
        self.update_screen_data()

    def _clear_screen_row(self, row_no):
        for widget in self.grid_slaves(row=row_no):
            widget.grid_remove()

    def update_screen_widgets(self, available_screen_height):
        max_screen_rows = available_screen_height // self.screen_row_height
        target_screen_row_count = max(
            min(
                max_screen_rows,
                self.header_row_count
                + self.data_row_count
                + self.footer_row_count
                - self.first_visible_data_row_no,
            ),
            self.header_row_count + 1 + self.footer_row_count,
        )
        # target_screen_row_count = 30

        # remove cells not required anymore ...
        while self.screen_row_count > target_screen_row_count:
            # print("removing")
            self._clear_screen_row(self.screen_row_count - 1)
            self.screen_row_count -= 1

        # ... or add cells that can be shown
        while self.screen_row_count < target_screen_row_count:
            # print("adding")
            for col in range(self.column_count):
                w = self.get_data_widget(self.screen_row_count, col)
                w.grid(
                    row=self.screen_row_count, column=col, sticky="nsew", pady=(0, 1), padx=(0, 1)
                )

            self.screen_row_count += 1

        self.visible_data_row_count = (
            self.screen_row_count - self.header_row_count - self.footer_row_count
        )

    def update_screen_data(self):
        self.update_screen_widgets(self.winfo_height())
        for screen_row_no in range(self.header_row_count, self.screen_row_count):
            data_row_no = self.first_visible_data_row_no + screen_row_no - self.header_row_count
            if data_row_no == self.data_row_count:
                break

            for col_no in range(self.column_count):
                w = self.get_data_widget(screen_row_no, col_no)
                value = self.get_data_value(data_row_no, col_no)
                if value is None:
                    w.configure(text="")
                else:
                    w.configure(text=str(value))

    def get_data_value(self, row_no, col_no):
        """ lazy dict version:
        assert 0 <= row_no < self.data_row_count
        if row_no in self.data_rows:
            return self.data_rows[row_no][col_no]
        else:
            return ""
        """
        return self.data_rows[row_no][col_no]

    def get_header_value(self, row_no, col_no):
        return self.header_rows[row_no][col_no]

    def on_configure(self, event):
        # query row height
        _, _, _, height = self.grid_bbox(row=1)
        if height > 10 and height < 100:
            "self.screen_row_height = height + 2"

        # screen_available_height = self.winfo_height()

        # print("HE", self.winfo_height(), event.height, self.screen_row_height)
        self.update_screen_widgets(event.height)

        self.update_screen_data()


class ScrollableGridTable(ttk.Frame):
    def __init__(self, master, header_rows, data_row_count, footer_row_count, frozen_column_count):
        ttk.Frame.__init__(self, master)

        # set up scrolling with canvas
        hscrollbar = ttk.Scrollbar(self, orient=tk.HORIZONTAL)
        self.canvas = tk.Canvas(self, bd=0, highlightthickness=0, xscrollcommand=hscrollbar.set)
        get_workbench().bind_all("<Control-r>", self.debug)
        self.create_infopanel(data_row_count)
        hscrollbar.config(command=self.canvas.xview)
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)

        self.canvas.grid(row=0, column=0, columnspan=2, sticky=tk.NSEW)
        self.infopanel.grid(row=1, column=0, sticky=tk.NSEW)
        hscrollbar.grid(row=1, column=1, sticky=tk.NSEW)

        # vertical scrollbar performs virtual scrolling
        self.vscrollbar = ttk.Scrollbar(
            self, orient=tk.VERTICAL, command=self._handle_vertical_scroll
        )
        self.vscrollbar.grid(row=0, column=2, sticky=tk.NSEW)

        self.columnconfigure(1, weight=1)
        self.rowconfigure(0, weight=1)

        self.interior = ttk.Frame(self.canvas)
        self.interior.columnconfigure(0, weight=1)
        self.interior.rowconfigure(0, weight=1)
        self.interior_id = self.canvas.create_window(0, 0, window=self.interior, anchor=tk.NW)
        self.bind("<Configure>", self._configure_interior, True)
        self.bind("<Expose>", self._on_expose, True)

        self.grid_table = GridTable(
            self.interior, header_rows, data_row_count, footer_row_count, frozen_column_count
        )
        self.grid_table.grid(row=0, column=0, sticky=tk.NSEW)

        self._update_vertical_scrollbar()

    def debug(self, event=None):
        print("DE", self.vscrollbar.get())

    def create_infopanel(self, data_row_count):
        self.infopanel = ttk.Frame(self)
        self.size_label = ttk.Label(self.infopanel, text=str(data_row_count) + " rows")
        self.size_label.grid(row=0, column=0, padx=5)

    def _update_vertical_scrollbar(self):
        first = self.grid_table.first_visible_data_row_no / self.grid_table.data_row_count
        last = first + self.grid_table.visible_data_row_count / self.grid_table.data_row_count
        # print(first, last, self.grid_table.visible_data_row_count)
        self.vscrollbar.set(first, last)

    def _handle_vertical_scroll(self, *args):
        # print("vscroll", args, self.vscrollbar.get())
        if len(args) == 3 and args[0] == "scroll":
            amount = int(args[1])
            unit = args[2]
            if unit == "pages":
                amount *= self.grid_table.visible_data_row_count

            self.grid_table.set_first_visible_data_row_no(
                self.grid_table.first_visible_data_row_no + amount
            )
        else:
            assert args[0] == "moveto"
            pos = max(min(float(args[1]), 1.0), 0.0)
            top_row = math.floor(
                (self.grid_table.data_row_count - self.grid_table.visible_data_row_count + 1) * pos
            )
            self.grid_table.set_first_visible_data_row_no(top_row)

        self._update_vertical_scrollbar()

    def _on_expose(self, event):
        self.update_idletasks()
        self._configure_interior(event)

    def _configure_interior(self, event):
        # update the scrollbars to match the size of the inner frame
        size = (self.interior.winfo_reqwidth(), self.canvas.winfo_height())
        self.canvas.config(scrollregion="0 0 %s %s" % size)
        if (
            self.interior.winfo_reqheight() != self.canvas.winfo_height()
            and self.canvas.winfo_height() > 10
        ):
            # update the interior's height to fit canvas
            self.canvas.itemconfigure(self.interior_id, height=self.canvas.winfo_height())

        self._update_vertical_scrollbar()



import tkinter as tk
from tkinter import ttk

from thonny import get_runner, get_workbench
from thonny.config_ui import ConfigurationPage
from thonny.ui_utils import create_string_var


class BackendDetailsConfigPage(ConfigurationPage):
    def should_restart(self):
        raise NotImplementedError()


class OnlyTextConfigurationPage(BackendDetailsConfigPage):
    def __init__(self, master, text):
        super().__init__(master)
        label = ttk.Label(self, text=text)
        label.grid()

    def should_restart(self):
        return False


class BackendConfigurationPage(ConfigurationPage):
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)

        self._backend_specs_by_desc = {
            spec.description: spec for spec in get_workbench().get_backends().values()
        }
        self._conf_pages = {}
        self._current_page = None

        current_backend_name = get_workbench().get_option("run.backend_name")
        try:
            current_backend_desc = get_workbench().get_backends()[current_backend_name].description
        except KeyError:
            current_backend_desc = ""

        self._combo_variable = create_string_var(current_backend_desc)

        label = ttk.Label(
            self, text=_("Which interpreter or device should Thonny use for running your code?")
        )
        label.grid(row=0, column=0, columnspan=2, sticky=tk.W)

        sorted_backend_specs = sorted(
            self._backend_specs_by_desc.values(), key=lambda x: x.sort_key
        )

        self._combo = ttk.Combobox(
            self,
            exportselection=False,
            textvariable=self._combo_variable,
            values=[spec.description for spec in sorted_backend_specs],
        )

        self._combo.grid(row=1, column=0, columnspan=2, sticky=tk.NSEW, pady=(0, 10))
        self._combo.state(["!disabled", "readonly"])

        self.labelframe = ttk.LabelFrame(self, text=" " + _("Details") + " ")
        self.labelframe.grid(row=2, column=0, sticky="nsew")
        self.labelframe.columnconfigure(0, weight=1)
        self.labelframe.rowconfigure(0, weight=1)

        self.columnconfigure(0, weight=1)
        self.rowconfigure(2, weight=1)

        self._combo_variable.trace("w", self._backend_changed)
        self._backend_changed()

    def _backend_changed(self, *args):
        backend_desc = self._combo_variable.get()

        if backend_desc == "":
            if self._current_page is not None:
                self._current_page.grid_forget()
            return

        page = self._get_conf_page(backend_desc)

        if page != self._current_page:

            if self._current_page is not None:
                self._current_page.grid_forget()

            page.grid(sticky="nsew", padx=10, pady=5)
            self._current_page = page

    def _get_conf_page(self, backend_desc):
        if backend_desc not in self._conf_pages:
            cp_constructor = self._backend_specs_by_desc[backend_desc].config_page_constructor
            if isinstance(cp_constructor, str):
                self._conf_pages[backend_desc] = OnlyTextConfigurationPage(
                    self.labelframe, cp_constructor
                )
            else:
                assert issubclass(cp_constructor, ConfigurationPage)
                self._conf_pages[backend_desc] = cp_constructor(self.labelframe)

        return self._conf_pages[backend_desc]

    def apply(self):
        if self._current_page is None:
            return None

        result = self._current_page.apply()

        if result is False:
            return False

        backend_desc = self._combo_variable.get()
        backend_name = self._backend_specs_by_desc[backend_desc].name
        get_workbench().set_option("run.backend_name", backend_name)

        if getattr(self._combo_variable, "modified") or self._current_page.should_restart():
            get_runner().restart_backend(False)

        return None


def load_plugin() -> None:
    def select_device():
        get_workbench().show_options("interpreter")

    get_workbench().add_configuration_page(
        "interpreter", _("Interpreter"), BackendConfigurationPage, 20
    )
    get_workbench().add_command(
        "select_interpreter", "run", _("Select interpreter") + "...", select_device, group=1
    )



# -*- coding: utf-8 -*-
import platform
import os
import sys

from thonny.code import get_saved_current_script_filename
from thonny import get_runner, terminal, get_workbench
from thonny.common import get_exe_dirs, get_augmented_system_path
from thonny.running import get_environment_overrides_for_python_subprocess


def _open_system_shell():
    """Main task is to modify path and open terminal window.
    Bonus (and most difficult) part is executing a script in this window
    for recommending commands for running given python and related pip"""

    cwd = get_workbench().get_local_cwd()

    proxy = get_runner().get_backend_proxy()
    if proxy and proxy.get_local_executable():
        target_executable = proxy.get_local_executable()
    else:
        target_executable = sys.executable

    exe_dirs = get_exe_dirs()
    if hasattr(proxy, "get_exe_dirs") and proxy.get_exe_dirs():
        # use both backend and frontend exe dirs
        exe_dirs = proxy.get_exe_dirs() + exe_dirs

    env_overrides = get_environment_overrides_for_python_subprocess(target_executable)
    env_overrides["PATH"] = get_augmented_system_path(exe_dirs)

    explainer = os.path.join(os.path.dirname(__file__), "explain_environment.py")
    cmd = [target_executable, explainer]

    activate = os.path.join(
        os.path.dirname(target_executable),
        "activate.bat" if platform.system() == "Windows" else "activate",
    )

    if os.path.isfile(activate):
        del env_overrides["PATH"]
        if platform.system() == "Windows":
            cmd = [activate, "&"] + cmd
        else:
            cmd = ["source", activate, ";"] + cmd

    return terminal.run_in_terminal(cmd, cwd, env_overrides, True)


def load_plugin() -> None:
    get_workbench().add_command(
        "OpenSystemShell",
        "tools",
        _("Open system shell..."),
        _open_system_shell,
        group=80,
        image="terminal",
    )



from typing import Callable, Dict, Union  # @UnusedImport

from thonny import get_workbench
from thonny.misc_utils import running_on_linux, running_on_windows, running_on_mac_os
from thonny.workbench import BasicUiThemeSettings, CompoundUiThemeSettings


def scale(value) -> float:
    # dimensions in this module were designed with a 1.67 scale
    return get_workbench().scale(value / 1.67)


def _treeview_settings() -> BasicUiThemeSettings:
    light_blue = "#ADD8E6"
    light_grey = "#D3D3D3"

    if running_on_linux() or running_on_mac_os():
        bg_sel_focus = light_blue
        fg_sel_focus = "black"
        fg_sel_notfocus = "black"
    else:
        bg_sel_focus = "SystemHighlight"
        fg_sel_focus = "SystemHighlightText"
        fg_sel_notfocus = "SystemWindowText"

    return {
        "Treeview": {
            "configure": {"font": "TreeviewFont"},
            "map": {
                "background": [
                    ("selected", "focus", bg_sel_focus),
                    ("selected", "!focus", light_grey),
                ],
                "foreground": [
                    ("selected", "focus", fg_sel_focus),
                    ("selected", "!focus", fg_sel_notfocus),
                ],
            },
            "layout": [
                # get rid of borders
                ("Treeview.treearea", {"sticky": "nswe"})
            ],
        },
        "treearea": {"configure": {"borderwidth": 0}},
    }


def _menubutton_settings() -> BasicUiThemeSettings:
    return {
        "TMenubutton": {
            "configure": {"padding": scale(14)},
            "layout": [
                ("Menubutton.dropdown", {"side": "right", "sticky": "ns"}),
                (
                    "Menubutton.button",
                    {
                        "children": [
                            # ('Menubutton.padding', {'children': [
                            ("Menubutton.label", {"sticky": ""})
                            # ], 'expand': '1', 'sticky': 'we'})
                        ],
                        "expand": "1",
                        "sticky": "nswe",
                    },
                ),
            ],
        }
    }


def _paned_window_settings() -> BasicUiThemeSettings:
    return {"Sash": {"configure": {"sashthickness": scale(10)}}}


def _menu_settings() -> BasicUiThemeSettings:
    return {"Menubar": {"configure": {"activeborderwidth": 0, "relief": "flat"}}}


def _text_settings() -> BasicUiThemeSettings:
    return {
        "Text": {
            "configure": {
                "background": "SystemWindow" if running_on_windows() else "white",
                "foreground": "SystemWindowText" if running_on_windows() else "black",
            }
        },
        "Syntax.Text": {"map": {"background": [("readonly", "Yellow")]}},
        "Gutter": {"configure": {"background": "#e0e0e0", "foreground": "#999999"}},
    }


def _label_settings() -> BasicUiThemeSettings:
    return {"Url.TLabel": {"configure": {"foreground": "DarkBlue"}}}


def _button_notebook_settings() -> BasicUiThemeSettings:
    # Adapted from https://github.com/python/cpython/blob/2.7/Demo/tkinter/ttk/notebook_closebtn.py
    return {
        "closebutton": {
            "element create": (
                "image",
                "img_close",
                ("active", "pressed", "!disabled", "img_close_active"),
                ("active", "!disabled", "img_close_active"),
                {"padding": scale(2), "sticky": ""},
            )
        },
        "ButtonNotebook.TNotebook.Tab": {
            "layout": [
                (
                    "Notebook.tab",
                    {
                        "sticky": "nswe",
                        "children": [
                            (
                                "Notebook.padding",
                                {
                                    "side": "top",
                                    "sticky": "nswe",
                                    "children": [
                                        (
                                            "Notebook.focus",
                                            {
                                                "side": "left",
                                                "sticky": "nswe",
                                                "children": [
                                                    (
                                                        "Notebook.label",
                                                        {"side": "left", "sticky": ""},
                                                    )
                                                ],
                                            },
                                        ),
                                        ("Notebook.closebutton", {"side": "right", "sticky": ""}),
                                    ],
                                },
                            )
                        ],
                    },
                )
            ]
        },
    }


def clam() -> BasicUiThemeSettings:
    # Transcribed from https://github.com/tcltk/tk/blob/master/library/ttk/clamTheme.tcl
    defaultfg = "#000000"
    disabledfg = "#999999"
    frame = "#dcdad5"
    window = "#ffffff"
    dark = "#cfcdc8"
    darker = "#bab5ab"
    darkest = "#9e9a91"
    lighter = "#eeebe7"
    selectbg = "#4a6984"
    selectfg = "#ffffff"
    altindicator = "#5895bc"
    disabledaltindicator = "#a0a0a0"

    return {
        ".": {
            "configure": {
                "background": frame,
                "foreground": defaultfg,
                "bordercolor": darkest,
                "darkcolor": dark,
                "lightcolor": lighter,
                "troughcolor": darker,
                "selectbackground": selectbg,
                "selectforeground": selectfg,
                "selectborderwidth": 0,
                "font": "TkDefaultFont",
            },
            "map": {
                "background": [("disabled", frame), ("active", lighter)],
                "foreground": [("disabled", disabledfg)],
                "selectbackground": [("!focus", darkest)],
                "selectforeground": [("!focus", "white")],
            },
        },
        "TButton": {
            "configure": {
                "anchor": "center",
                "width": scale(11),
                "padding": scale(5),
                "relief": "raised",
            },
            "map": {
                "background": [("disabled", frame), ("pressed", darker), ("active", lighter)],
                "lightcolor": [("pressed", darker)],
                "darkcolor": [("pressed", darker)],
                "bordercolor": [("alternate", "#000000")],
            },
        },
        "Toolbutton": {
            "configure": {"anchor": "center", "padding": scale(2), "relief": "flat"},
            "map": {
                "relief": [
                    ("disabled", "flat"),
                    ("selected", "sunken"),
                    ("pressed", "sunken"),
                    ("active", "raised"),
                ],
                "background": [("disabled", frame), ("pressed", darker), ("active", lighter)],
                "lightcolor": [("pressed", darker)],
                "darkcolor": [("pressed", darker)],
            },
        },
        "TCheckbutton": {
            "configure": {
                "indicatorbackground": "#ffffff",
                "indicatormargin": [scale(1), scale(1), scale(6), scale(1)],
                "padding": scale(2),
            },
            "map": {
                "indicatorbackground": [
                    ("pressed", frame),
                    ("!disabled", "alternate", altindicator),
                    ("disabled", "alternate", disabledaltindicator),
                    ("disabled", frame),
                ]
            },
        },
        # TRadiobutton has same style as TCheckbutton
        "TRadiobutton": {
            "configure": {
                "indicatorbackground": "#ffffff",
                "indicatormargin": [scale(1), scale(1), scale(6), scale(1)],
                "padding": scale(2),
            },
            "map": {
                "indicatorbackground": [
                    ("pressed", frame),
                    ("!disabled", "alternate", altindicator),
                    ("disabled", "alternate", disabledaltindicator),
                    ("disabled", frame),
                ]
            },
        },
        "TMenubutton": {"configure": {"width": scale(11), "padding": scale(5), "relief": "raised"}},
        "TEntry": {
            "configure": {"padding": scale(1), "insertwidth": scale(1)},
            "map": {
                "background": [("readonly", frame)],
                "bordercolor": [("focus", selectbg)],
                "lightcolor": [("focus", "#6f9dc6")],
                "darkcolor": [("focus", "#6f9dc6")],
            },
        },
        "TCombobox": {
            "configure": {
                "padding": [scale(4), scale(2), scale(2), scale(2)],
                "insertwidth": scale(1),
            },
            "map": {
                "background": [("active", lighter), ("pressed", lighter)],
                "fieldbackground": [("readonly", "focus", selectbg), ("readonly", frame)],
                "foreground": [("readonly", "focus", selectfg)],
                "arrowcolor": [("disabled", disabledfg)],
            },
        },
        "ComboboxPopdownFrame": {"configure": {"relief": "solid", "borderwidth": scale(1)}},
        "TSpinbox": {
            "configure": {"arrowsize": scale(10), "padding": [scale(2), 0, scale(10), 0]},
            "map": {"background": [("readonly", frame)], "arrowcolor": [("disabled", disabledfg)]},
        },
        "TNotebook.Tab": {
            "configure": {"padding": [scale(6), scale(2), scale(6), scale(2)]},
            "map": {
                "padding": [("selected", [scale(6), scale(4), scale(6), scale(4)])],
                "background": [("selected", frame), ("", darker)],
                "lightcolor": [("selected", lighter), ("", dark)],
            },
        },
        "Treeview": {
            "configure": {"background": window},
            "map": {
                "background": [
                    ("disabled", frame),
                    ("!disabled", "!selected", window),
                    ("selected", selectbg),
                ],
                "foreground": [
                    ("disabled", disabledfg),
                    ("!disabled", "!selected", defaultfg),
                    ("selected", selectfg),
                ],
            },
        },
        # Treeview heading
        "Heading": {
            "configure": {
                "font": "TkHeadingFont",
                "relief": "raised",
                "padding": [scale(3), scale(3), scale(3), scale(3)],
            }
        },
        "TLabelframe": {"configure": {"labeloutside": True, "labelmargins": [0, 0, 0, scale(4)]}},
        "TProgressbar": {"configure": {"background": frame}},
        "Sash": {"configure": {"sashthickness": scale(6), "gripcount": 10}},
    }


def xpnative() -> BasicUiThemeSettings:
    # Transcribed from https://github.com/tcltk/tk/blob/master/library/ttk/xpTheme.tcl
    return {
        ".": {
            "configure": {
                "background": "SystemButtonFace",
                "foreground": "SystemWindowText",
                "selectbackground": "SystemHighlight",
                "selectforeground": "SystemHighlightText",
                "font": "TkDefaultFont",
            },
            "map": {"foreground": [("disabled", "SystemGrayText")]},
        },
        "TButton": {
            "configure": {"anchor": "center", "width": scale(11), "padding": [scale(1), scale(1)]}
        },
        "Toolbutton": {"configure": {"padding": [scale(4), scale(4)]}},
        "TCheckbutton": {"configure": {"padding": scale(2)}},
        # TRadiobutton has same style as TCheckbutton
        "TRadiobutton": {"configure": {"padding": scale(2)}},
        "TMenubutton": {"configure": {"padding": [scale(8), scale(4)]}},
        "TEntry": {
            "configure": {"padding": [scale(2), scale(2), scale(2), scale(4)]},
            "map": {
                "selectbackground": [("!focus", "SystemWindow")],
                "selectforeground": [("!focus", "SystemWindowText")],
            },
        },
        "TCombobox": {
            "configure": {"padding": scale(2)},
            "map": {
                "selectbackground": [("!focus", "SystemWindow")],
                "selectforeground": [("!focus", "SystemWindowText")],
                "foreground": [
                    ("disabled", "SystemGrayText"),
                    ("readonly", "focus", "SystemHighlightText"),
                ],
                "focusfill": [("readonly", "focus", "SystemHighlight")],
            },
        },
        "ComboboxPopdownFrame": {"configure": {"relief": "solid", "borderwidth": scale(1)}},
        "TSpinbox": {
            "configure": {"padding": [scale(2), 0, scale(14), 0]},
            "map": {
                "selectbackground": [("!focus", "SystemWindow")],
                "selectforeground": [("!focus", "SystemWindowText")],
            },
        },
        "TNotebook": {"configure": {"tabmargins": [scale(2), scale(2), scale(2), 0]}},
        "TNotebook.Tab": {
            "map": {"expand": [("selected", [scale(2), scale(2), scale(2), scale(2)])]}
        },
        "Treeview": {
            "configure": {"background": "SystemWindow"},
            "map": {
                "background": [
                    ("disabled", "SystemButtonFace"),
                    ("!disabled", "!selected", "SystemWindow"),
                    ("selected", "SystemHighlight"),
                ],
                "foreground": [
                    ("disabled", "SystemGrayText"),
                    ("!disabled", "!selected", "SystemWindowText"),
                    ("selected", "SystemHighlightText"),
                ],
            },
        },
        "Heading": {"configure": {"font": "TkHeadingFont", "relief": "raised"}},  # Treeview heading
        "TLabelframe.Label": {"configure": {"foreground": "#0046d5"}},
    }


def aqua() -> BasicUiThemeSettings:
    # https://github.com/tcltk/tk/blob/master/library/ttk/aquaTheme.tcl
    return {
        ".": {
            "configure": {
                "font": "TkDefaultFont",
                "background": "systemWindowBody",
                "foreground": "systemModelessDialogActiveText",
                "selectbackground": "systemHighlight",
                "selectforeground": "systemModelessDialogActiveText",
                "selectborderwidth": 0,
                "insertwidth": 1,
                "stipple": "",
            },
            "map": {
                "foreground": [
                    ("disabled", "systemModelessDialogInactiveText"),
                    ("background", "systemModelessDialogInactiveText"),
                ],
                "selectbackground": [
                    ("background", "systemHighlightSecondary"),
                    ("!focus", "systemHighlightSecondary"),
                ],
                "selectforeground": [
                    ("background", "systemModelessDialogInactiveText"),
                    ("!focus", "systemDialogActiveText"),
                ],
            },
        },
        "TButton": {"configure": {"anchor": "center", "width": "6"}},
        "Toolbutton": {"configure": {"padding": 4}},
        "TNotebook": {
            "configure": {"tabmargins": [10, 0], "tabposition": "n", "padding": [18, 8, 18, 17]}
        },
        "TNotebook.Tab": {"configure": {"padding": [12, 3, 12, 2]}},
        "TCombobox": {"configure": {"postoffset": [5, -2, -10, 0]}},
        "Heading": {"configure": {"font": "TkHeadingFont"}},
        "Treeview": {
            "configure": {"rowheight": 18, "background": "white"},
            "map": {
                "background": [
                    ("disabled", "systemDialogBackgroundInactive"),
                    ("!disabled", "!selected", "systemWindowBody"),
                    ("selected", "background", "systemHighlightSecondary"),
                    ("selected", "systemHighlight"),
                ],
                "foreground": [
                    ("disabled", "systemModelessDialogInactiveText"),
                    ("!disabled", "!selected", "black"),
                    ("selected", "systemModelessDialogActiveText"),
                ],
            },
        },
        "TProgressbar": {"configure": {"period": 100, "maxphase": 255}},
        "Labelframe": {"configure": {"labeloutside": True, "labelmargins": [14, 0, 14, 4]}},
    }


def windows() -> CompoundUiThemeSettings:
    return [
        xpnative(),
        _treeview_settings(),
        _menubutton_settings(),
        _paned_window_settings(),
        _menu_settings(),
        _text_settings(),
        _label_settings(),
        _button_notebook_settings(),
        {
            "TNotebook": {
                "configure": {
                    # With tabmargins I can get a gray line below tab, which separates
                    # tab content from label
                    "tabmargins": [scale(2), scale(2), scale(2), scale(2)]
                }
            },
            "Tab": {"configure": {"padding": [scale(3), scale(1), scale(3), 0]}},
            "ButtonNotebook.TNotebook.Tab": {
                "configure": {"padding": (scale(4), scale(1), scale(1), 0)}
            },
            "TCombobox": {
                "map": {
                    "selectbackground": [
                        ("readonly", "!focus", "SystemWindow"),
                        ("readonly", "focus", "SystemHighlight"),
                    ],
                    "selectforeground": [
                        ("readonly", "!focus", "SystemWindowText"),
                        ("readonly", "focus", "SystemHighlightText"),
                    ],
                }
            },
            "Listbox": {
                "configure": {
                    "background": "SystemWindow",
                    "foreground": "SystemWindowText",
                    "disabledforeground": "SystemGrayText",
                    "highlightbackground": "SystemActiveBorder",
                    "highlightcolor": "SystemActiveBorder",
                    "highlightthickness": scale(1),
                }
            },
            "ViewBody.TFrame": {
                "configure": {
                    "background": "SystemButtonFace"  # to create the fine line below toolbar
                }
            },
            "ViewToolbar.TFrame": {"configure": {"background": "SystemWindow"}},
            "ViewToolbar.Toolbutton": {"configure": {"background": "SystemWindow"}},
            "ViewTab.TLabel": {
                "configure": {"background": "SystemWindow", "padding": [scale(5), 0]}
            },
            "ViewToolbar.TLabel": {
                "configure": {"background": "SystemWindow", "padding": [scale(5), 0]}
            },
            "ViewToolbar.Link.TLabel": {
                "configure": {"background": "SystemWindow", "padding": [scale(5), 0]}
            },
            "Active.ViewTab.TLabel": {
                "configure": {
                    # "font" : "BoldTkDefaultFont",
                    "relief": "sunken",
                    "borderwidth": scale(1),
                }
            },
            "Inactive.ViewTab.TLabel": {"configure": {"font": "UnderlineTkDefaultFont"}},
        },
    ]


def enhanced_clam() -> CompoundUiThemeSettings:
    return [
        clam(),
        _treeview_settings(),
        _menubutton_settings(),
        _paned_window_settings(),
        _menu_settings(),
        _text_settings(),
        _label_settings(),
        _button_notebook_settings(),
        {
            "ButtonNotebook.Tab": {
                "configure": {"padding": (scale(6), scale(4), scale(2), scale(3))}
            },
            "TScrollbar": {
                "configure": {
                    "gripcount": 0,
                    "arrowsize": scale(14),
                    # "arrowcolor" : "DarkGray"
                    # "width" : 99 # no effect
                }
            },
            "TCombobox": {
                "configure": {"arrowsize": scale(14)},
                "map": {
                    "selectbackground": [("readonly", "!focus", "#dcdad5")],
                    "selectforeground": [("readonly", "!focus", "#000000")],
                },
            },
            "TCheckbutton": {"configure": {"indicatorsize": scale(12)}},
            "TRadiobutton": {"configure": {"indicatorsize": scale(12)}},
            "Listbox": {
                "configure": {
                    "background": "white",
                    "foreground": "black",
                    "disabledforeground": "#999999",
                    "highlightbackground": "#4a6984",
                    "highlightcolor": "#4a6984",
                    "highlightthickness": scale(1),
                }
            },
        },
    ]


def enhanced_aqua() -> CompoundUiThemeSettings:
    return [
        _treeview_settings(),
        _menubutton_settings(),
        # _paned_window_settings(),
        _menu_settings(),
        {
            "TPanedWindow": {"configure": {"background": "systemDialogBackgroundActive"}},
            "TFrame": {"configure": {"background": "systemDialogBackgroundActive"}},
            "Tab": {"map": {"foreground": [("selected", "white")]}},
        },
    ]


def load_plugin() -> None:
    from tkinter import ttk

    original_themes = ttk.Style().theme_names()

    # load all base themes
    for name in original_themes:
        settings = {}  # type: Union[Dict, Callable[[], Dict]]
        if name == "clam":
            settings = clam
        elif name == "xpnative":
            settings = xpnative
        elif name == "aqua":
            settings = aqua

        get_workbench().add_ui_theme(name, None, settings)

    get_workbench().add_ui_theme(
        "Enhanced Clam",
        "clam",
        enhanced_clam,
        {"tab-close": "tab-close-clam", "tab-close-active": "tab-close-active-clam"},
    )

    if "xpnative" in original_themes:
        get_workbench().add_ui_theme("Windows", "xpnative", windows)

    if "aqua" in original_themes:
        get_workbench().add_ui_theme("Kind of Aqua", "aqua", enhanced_aqua)



from thonny import get_workbench

BASE_LANGUAGE_CODE = "en_US"
BASE_LANGUAGE_NAME = "English"

# http://www.internationalphoneticalphabet.org/languages/language-names-in-native-language/

LANGUAGES_DICT = {
    "de_DE": "Deutsch [ALPHA]",
    "et_EE": "Eesti",
    BASE_LANGUAGE_CODE: BASE_LANGUAGE_NAME,
    "es_ES": "Español [ALPHA]",
    "fr_FR": "Français [BETA]",
    "it_IT": "Italiano [ALPHA]",
    "nl_NL": "Nederlands [BETA]",
    "pl_PL": "Polski [BETA]",
    "pt_PT": "Português (PT) [BETA]",
    "pt_BR": "Português (BR) [ALPHA]",
    "ru_RU": "Русский",
    "tr_TR": "Türkçe [BETA]",
    "uk_UA": "Українська",
    "zh_TW": "繁體中文-TW",
    "zh_CN": "简体中文 ",
    "ja_JP": "日本語  [ALPHA]",
    "lt_LT": "Lietuvių",
    "el_GR": "Ελληνικά [BETA]",
}

# how many spaces to add to button caption in order to make whole text visible
BUTTON_PADDING_SIZES = {"zh_TW": 4, "zh_CN": 4, "ja_JP": 4}


def get_button_padding():
    code = get_workbench().get_option("general.language")
    if code in BUTTON_PADDING_SIZES:
        return BUTTON_PADDING_SIZES[code] * " "
    else:
        return ""


def get_language_code_by_name(name):
    for code in LANGUAGES_DICT:
        if LANGUAGES_DICT[code] == name:
            return code



class I2C:
    ""

    def deinit():
        pass

    def readfrom_into():
        pass

    def scan():
        pass

    def try_lock():
        pass

    def unlock():
        pass

    def writeto():
        pass


class OneWire:
    ""

    def deinit():
        pass

    def read_bit():
        pass

    def reset():
        pass

    def write_bit():
        pass


class SPI:
    ""

    def configure():
        pass

    def deinit():
        pass

    frequency = None

    def readinto():
        pass

    def try_lock():
        pass

    def unlock():
        pass

    def write():
        pass

    def write_readinto():
        pass


class UART:
    ""
    Parity = None
    baudrate = None

    def deinit():
        pass

    in_waiting = None

    def read():
        pass

    def readinto():
        pass

    def readline():
        pass

    def reset_input_buffer():
        pass

    def write():
        pass



class ArithmeticError:
    ""


class AssertionError:
    ""


class AttributeError:
    ""


class BaseException:
    ""


class EOFError:
    ""


Ellipsis = None


class Exception:
    ""


class GeneratorExit:
    ""


class ImportError:
    ""


class IndentationError:
    ""


class IndexError:
    ""


class KeyError:
    ""


class KeyboardInterrupt:
    ""


class LookupError:
    ""


class MemoryError:
    ""


class NameError:
    ""


class NotImplementedError:
    ""


class OSError:
    ""


class OverflowError:
    ""


class RuntimeError:
    ""


class StopIteration:
    ""


class SyntaxError:
    ""


class SystemExit:
    ""


class TypeError:
    ""


class UnicodeError:
    ""


class ValueError:
    ""


class ZeroDivisionError:
    ""


def abs():
    pass


def all():
    pass


def any():
    pass


def bin():
    pass


class bool:
    ""


class bytearray:
    ""

    def append():
        pass

    def extend():
        pass


class bytes:
    ""

    def count():
        pass

    def endswith():
        pass

    def find():
        pass

    def format():
        pass

    def index():
        pass

    def isalpha():
        pass

    def isdigit():
        pass

    def islower():
        pass

    def isspace():
        pass

    def isupper():
        pass

    def join():
        pass

    def lower():
        pass

    def lstrip():
        pass

    def replace():
        pass

    def rfind():
        pass

    def rindex():
        pass

    def rsplit():
        pass

    def rstrip():
        pass

    def split():
        pass

    def startswith():
        pass

    def strip():
        pass

    def upper():
        pass


def callable():
    pass


def chr():
    pass


class classmethod:
    ""


class dict:
    ""

    def clear():
        pass

    def copy():
        pass

    def fromkeys():
        pass

    def get():
        pass

    def items():
        pass

    def keys():
        pass

    def pop():
        pass

    def popitem():
        pass

    def setdefault():
        pass

    def update():
        pass

    def values():
        pass


def dir():
    pass


def divmod():
    pass


class enumerate:
    ""


def eval():
    pass


def exec():
    pass


class filter:
    ""


class float:
    ""


class frozenset:
    ""

    def add():
        pass

    def clear():
        pass

    def copy():
        pass

    def difference():
        pass

    def difference_update():
        pass

    def discard():
        pass

    def intersection():
        pass

    def intersection_update():
        pass

    def isdisjoint():
        pass

    def issubset():
        pass

    def issuperset():
        pass

    def pop():
        pass

    def remove():
        pass

    def symmetric_difference():
        pass

    def symmetric_difference_update():
        pass

    def union():
        pass

    def update():
        pass


def getattr():
    pass


def globals():
    pass


def hasattr():
    pass


def hash():
    pass


def help():
    pass


def hex():
    pass


def id():
    pass


def input():
    pass


class int:
    ""

    def from_bytes():
        pass

    def to_bytes():
        pass


def isinstance():
    pass


def issubclass():
    pass


def iter():
    pass


def len():
    pass


class list:
    ""

    def append():
        pass

    def clear():
        pass

    def copy():
        pass

    def count():
        pass

    def extend():
        pass

    def index():
        pass

    def insert():
        pass

    def pop():
        pass

    def remove():
        pass

    def reverse():
        pass

    def sort():
        pass


def locals():
    pass


class map:
    ""


def max():
    pass


def min():
    pass


def next():
    pass


class object:
    ""


def oct():
    pass


def open():
    pass


def ord():
    pass


def pow():
    pass


def print():
    pass


class range:
    ""


def repr():
    pass


class reversed:
    ""


def round():
    pass


class set:
    ""

    def add():
        pass

    def clear():
        pass

    def copy():
        pass

    def difference():
        pass

    def difference_update():
        pass

    def discard():
        pass

    def intersection():
        pass

    def intersection_update():
        pass

    def isdisjoint():
        pass

    def issubset():
        pass

    def issuperset():
        pass

    def pop():
        pass

    def remove():
        pass

    def symmetric_difference():
        pass

    def symmetric_difference_update():
        pass

    def union():
        pass

    def update():
        pass


def setattr():
    pass


class slice:
    ""


def sorted():
    pass


class staticmethod:
    ""


class str:
    ""

    def count():
        pass

    def endswith():
        pass

    def find():
        pass

    def format():
        pass

    def index():
        pass

    def isalpha():
        pass

    def isdigit():
        pass

    def islower():
        pass

    def isspace():
        pass

    def isupper():
        pass

    def join():
        pass

    def lower():
        pass

    def lstrip():
        pass

    def replace():
        pass

    def rfind():
        pass

    def rindex():
        pass

    def rsplit():
        pass

    def rstrip():
        pass

    def split():
        pass

    def startswith():
        pass

    def strip():
        pass

    def upper():
        pass


def sum():
    pass


class super:
    ""


class tuple:
    ""

    def count():
        pass

    def index():
        pass


class type:
    ""


class zip:
    ""



import tkinter as tk
from tkinter import ttk
from thonny.ui_utils import CommonDialog

page_specs = []


class ExportDialog(CommonDialog):
    def __init__(self, master):
        super().__init__(master=master)

        self.title("Export file")
        self.protocol("WM_DELETE_WINDOW", self.on_cancel)

        mainframe = ttk.Frame(self)
        mainframe.grid(row=0, column=0, sticky="nsew")
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.notebook = ttk.Notebook(mainframe)
        self.notebook.grid(row=3, column=0, columnspan=3, padx=20, pady=(20, 0), sticky="nsew")

        self.ok_button = ttk.Button(mainframe, text="OK", command=self.on_ok)
        self.ok_button.grid(row=4, column=1, sticky="e", padx=(20, 10), pady=(10, 20))

        self.cancel_button = ttk.Button(mainframe, text="Cancel", command=self.on_cancel)
        self.cancel_button.grid(row=4, column=2, sticky="e", padx=(0, 20), pady=(10, 20))

        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(3, weight=1)

        for title, page_class in page_specs:
            page = page_class(self.notebook)
            self.notebook.add(page, text=title)

    def on_ok(self):
        print("OK")

    def on_cancel(self):
        print("Cancel")
        self.destroy()


class LocalFilesExportPage(ttk.Frame):
    def __init__(self, master):
        super().__init__(master=master)


def add_export_page(title, page_class):
    page_specs.append((title, page_class))



from thonny.common import (
    InputSubmission,
    InterruptCommand,
    EOFCommand,
    parse_message,
    ToplevelCommand,
    ToplevelResponse,
    InlineCommand,
    InlineResponse,
    UserError,
    serialize_message,
    BackendEvent,
    ValueInfo,
    execute_system_command,
)
import sys
import logging
import traceback
import queue
from thonny.plugins.micropython.connection import (
    ConnectionClosedException,
    ConnectionFailedException,
)
from textwrap import dedent
import ast
import re
from queue import Queue, Empty
import threading
import os
import time
from thonny.misc_utils import find_volumes_by_name, sizeof_fmt
import jedi
import io
import tokenize
from thonny.running import EXPECTED_TERMINATION_CODE
import binascii
import shutil

# See https://github.com/dhylands/rshell/blob/master/rshell/main.py
# for UART_BUFFER_SIZE vs USB_BUFFER_SIZE
# ampy uses 32 bytes: https://github.com/pycampers/ampy/blob/master/ampy/files.py
# I'm not worrying so much, because reader thread reads continuously
# and writer (SerialConnection) has it's own blocks and delays
BUFFER_SIZE = 512

BAUDRATE = 115200
ENCODING = "utf-8"

# Commands
RAW_MODE_CMD = b"\x01"
NORMAL_MODE_CMD = b"\x02"
INTERRUPT_CMD = b"\x03"
SOFT_REBOOT_CMD = b"\x04"

# Output tokens
THONNY_MSG_START = b"\x02<thonny>"
THONNY_MSG_END = b"</thonny>\x04"
EOT = b"\x04"
NORMAL_PROMPT = b">>> "
LF = b"\n"
OK = b"OK"

# first prompt when switching to raw mode (or after soft reboot in raw mode)
# Looks like it's not translatable in CP
# https://github.com/adafruit/circuitpython/blob/master/locale/circuitpython.pot
FIRST_RAW_PROMPT = b"raw REPL; CTRL-B to exit\r\n>"
FIRST_RAW_PROMPT_SUFFIX = b"\r\n>"

RAW_PROMPT = b">"

BLOCK_CLOSERS = re.compile(
    b"|".join(map(re.escape, [LF, EOT, THONNY_MSG_START, NORMAL_PROMPT, FIRST_RAW_PROMPT]))
)


logger = logging.getLogger("thonny.micropython.backend")


def debug(msg):
    return
    print(msg, file=sys.stderr)


class MicroPythonBackend:
    def __init__(self, connection, clean, api_stubs_path):
        self._connection = connection
        self._local_cwd = None
        self._cwd = None
        self._interrupt_requested = False
        self._cancel_requested = False
        self._command_queue = Queue()  # populated by reader thread
        self._progress_times = {}

        self._api_stubs_path = api_stubs_path

        self._command_reading_thread = threading.Thread(target=self._read_commands, daemon=True)
        self._command_reading_thread.start()

        self._startup_time = time.time()
        self._ctrl_suggestion_given = False

        try:
            self._prepare(clean)
            self._mainloop()
        except ConnectionClosedException as e:
            self._on_connection_closed(e)
        except Exception:
            logger.exception("Crash in backend")
            traceback.print_exc()

    def _prepare(self, clean):
        if clean:
            self._interrupt_to_raw_prompt()
            self._clear_environment()
        else:
            self._process_until_initial_raw_prompt()

        self._cwd = self._fetch_cwd()
        self._welcome_text = self._fetch_welcome_text()
        self._builtin_modules = self._fetch_builtin_modules()
        self._builtins_info = self._fetch_builtins_info()

        self._send_ready_message()

    def _mainloop(self):
        while True:
            try:
                self._cancel_requested = False
                self._interrupt_requested = False
                self._check_for_connection_errors()
                try:
                    cmd = self._command_queue.get(timeout=0.1)
                except Empty:
                    # No command in queue, but maybe a thread produced output meanwhile
                    # or the user resetted the device
                    self._forward_unexpected_output()
                    continue

                if isinstance(cmd, InputSubmission):
                    self._submit_input(cmd.data)
                elif isinstance(cmd, EOFCommand):
                    self._soft_reboot(False)
                elif isinstance(cmd, InterruptCommand):
                    self._interrupt()
                else:
                    self.handle_command(cmd)
            except KeyboardInterrupt:
                self._interrupt()

    def _fetch_welcome_text(self):
        self._connection.write(NORMAL_MODE_CMD)
        welcome_text = self._connection.read_until(NORMAL_PROMPT).strip(b"\r\n >")
        if os.name != "nt":
            welcome_text = welcome_text.replace(b"\r\n", b"\n")

        # Go back to raw prompt
        self._connection.write(RAW_MODE_CMD)
        self._connection.read_until(FIRST_RAW_PROMPT)

        return welcome_text.decode(ENCODING, errors="replace")

    def _fetch_uname(self):
        res = self._evaluate("__thonny_os.uname()", prelude="import os as __thonny_os")
        return {
            "sysname": res[0],
            "nodename": res[1],
            "release": res[2],
            "version": res[3],
            "machine": res[4],
        }

    def _fetch_builtin_modules(self):
        out, err, _ = self._execute("help('modules')", capture_output=True)
        assert not err, "Error was: %r" % err

        modules_str_lines = out.strip().splitlines()

        last_line = modules_str_lines[-1].strip()
        if last_line.count(" ") > 0 and "  " not in last_line and "\t" not in last_line:
            # probably something like "plus any modules on the filesystem"
            # (can be in different languages)
            modules_str_lines = modules_str_lines[:-1]

        modules_str = (
            " ".join(modules_str_lines)
            .replace("/__init__", "")
            .replace("__main__", "")
            .replace("/", ".")
        )

        return modules_str.split()

    def _fetch_builtins_info(self):
        """
        for p in self._get_api_stubs_path():
            builtins_file = os.path.join(p, "__builtins__.py")
            if os.path.exists(builtins_file):
                return parse_api_information(builtins_file)
        """
        path = os.path.join(self._api_stubs_path, "builtins.py")
        if os.path.exists(path):
            return parse_api_information(path)
        else:
            return {}

    def _fetch_cwd(self):
        return self._evaluate(
            "__thonny_os.getcwd() if hasattr(__thonny_os, 'getcwd') else ''",
            prelude="import os as __thonny_os",
        )

    def _send_ready_message(self):
        self.send_message(ToplevelResponse(welcome_text=self._welcome_text, cwd=self._cwd))

    def _check_send_inline_progress(self, cmd, value, maximum, description=None):
        assert "id" in cmd
        prev_time = self._progress_times.get(cmd["id"], 0)
        if value != maximum and time.time() - prev_time < 0.2:
            # Don't notify too often
            return
        else:
            self._progress_times[cmd["id"]] = time.time()

        if description is None:
            description = cmd.get("description", "Working...")

        self.send_message(
            BackendEvent(
                event_type="InlineProgress",
                command_id=cmd["id"],
                value=value,
                maximum=maximum,
                description=description,
            )
        )

    def _interrupt(self):
        self._connection.write(INTERRUPT_CMD)

    def _check_for_interrupt(self, action_scope):
        if action_scope == "device" and self._interrupt_requested:
            self._interrupt()
            self._interrupt_requested = False

        if action_scope == "local" and self._cancel_requested:
            self._cancel_requested = False
            raise KeyboardInterrupt()

    def _interrupt_to_raw_prompt(self):
        # NB! Sometimes disconnecting and reconnecting (on macOS?)
        # too quickly causes anomalies. See CalliopeMiniProxy for more details

        discarded_bytes = b""

        for delay in [0.05, 0.5, 0.1, 1.0, 3.0, 5.0]:
            # Interrupt several times, because with some drivers first interrupts seem to vanish
            if delay >= 1:
                self._show_error(
                    "Could not enter REPL. Trying again with %d second waiting time..." % delay
                )
            self._connection.reset_output_buffer()
            self._connection.write(INTERRUPT_CMD)
            self._connection.write(RAW_MODE_CMD)
            time.sleep(delay)
            discarded_bytes += self._connection.read_all()
            if discarded_bytes.endswith(FIRST_RAW_PROMPT) or discarded_bytes.endswith(b"\r\n>"):
                break
        else:
            max_tail_length = 500
            if len(discarded_bytes) > max_tail_length:
                discarded_bytes_str = (
                    "[skipping %d bytes] ..." % (len(discarded_bytes) - max_tail_length)
                ) + repr(discarded_bytes[:-max_tail_length])
            else:
                discarded_bytes_str = repr(discarded_bytes)
            self._show_error(
                "Could not enter REPL. Giving up. Read bytes:\n"
                + discarded_bytes_str
                + "\n\nYour options:\n\n"
                + "  - check connection properties;\n"
                + "  - make sure the device has suitable firmware;\n"
                + "  - make sure the device is not in bootloader mode;\n"
                + "  - reset the device and try again;\n"
                + "  - try other serial clients (Putty, TeraTerm, screen, ...);\n"
                + "  - ask for help in Thonny's forum or issue tracker."
            )
            sys.exit()

    def _soft_reboot(self, side_command):
        if side_command:
            self._interrupt_to_raw_prompt()

        # Need to go to normal mode. MP doesn't run user code in raw mode
        # (CP does, but it doesn't hurt to do it there as well)
        self._connection.write(NORMAL_MODE_CMD)
        self._connection.read_until(NORMAL_PROMPT)

        self._connection.write(SOFT_REBOOT_CMD)

        if not side_command:
            self._process_until_raw_prompt()
            self.send_message(ToplevelResponse(cwd=self._cwd))

    def _read_commands(self):
        "works in separate thread"

        while True:
            line = sys.stdin.readline()
            if line == "":
                logger.info("Read stdin EOF")
                sys.exit()
            cmd = parse_message(line)
            if isinstance(cmd, InterruptCommand):
                self._interrupt_requested = True
                self._cancel_requested = True
            else:
                self._command_queue.put(cmd)

    def handle_command(self, cmd):
        assert isinstance(cmd, (ToplevelCommand, InlineCommand))

        if "local_cwd" in cmd:
            self._local_cwd = cmd["local_cwd"]

        def create_error_response(**kw):
            if not "error" in kw:
                kw["error"] = traceback.format_exc()

            if isinstance(cmd, ToplevelCommand):
                return ToplevelResponse(command_name=cmd.name, **kw)
            else:
                return InlineResponse(command_name=cmd.name, **kw)

        handler = getattr(self, "_cmd_" + cmd.name, None)

        if handler is None:
            response = create_error_response(error="Unknown command: " + cmd.name)
        else:
            try:
                response = handler(cmd)
            except SystemExit:
                # Must be caused by Thonny or plugins code
                if isinstance(cmd, ToplevelCommand):
                    traceback.print_exc()
                response = create_error_response(SystemExit=True)
            except UserError as e:
                sys.stderr.write(str(e) + "\n")
                response = create_error_response()
            except KeyboardInterrupt:
                response = create_error_response(error="Interrupted", interrupted=True)
            except Exception:
                _report_internal_error()
                response = create_error_response(context_info="other unhandled exception")

        if response is None:
            response = {}

        if response is False:
            # Command doesn't want to send any response
            return

        elif isinstance(response, dict):
            if isinstance(cmd, ToplevelCommand):
                response = ToplevelResponse(command_name=cmd.name, **response)
            elif isinstance(cmd, InlineCommand):
                response = InlineResponse(cmd.name, **response)

        if "id" in cmd and "command_id" not in response:
            response["command_id"] = cmd["id"]

        debug("cmd: " + str(cmd) + ", respin: " + str(response))
        self.send_message(response)

    def _submit_input(self, cdata: str) -> None:
        # TODO: what if there is a previous unused data waiting
        assert self._connection.outgoing_is_empty()

        assert cdata.endswith("\n")
        if not cdata.endswith("\r\n"):
            # submission is done with CRLF
            cdata = cdata[:-1] + "\r\n"

        bdata = cdata.encode(ENCODING)

        self._connection.write(bdata)
        # Try to consume the echo

        try:
            echo = self._connection.read(len(bdata))
        except queue.Empty:
            # leave it.
            logging.warning("Timeout when reading echo")
            return

        if echo != bdata:
            # because of autoreload? timing problems? interruption?
            # Leave it.
            logging.warning("Unexpected echo. Expected %s, got %s" % (bdata, echo))
            self._connection.unread(echo)

    def send_message(self, msg):
        if "cwd" not in msg:
            msg["cwd"] = self._cwd

        sys.stdout.write(serialize_message(msg) + "\n")
        sys.stdout.flush()

    def _send_output(self, data, stream_name):
        if not data:
            return
        data = self._transform_output(data)
        msg = BackendEvent(event_type="ProgramOutput", stream_name=stream_name, data=data)
        self.send_message(msg)

    def _transform_output(self, data):
        # Any keypress wouldn't work
        return data.replace(
            "Press any key to enter the REPL. Use CTRL-D to reload.",
            "Press Ctrl-C to enter the REPL. Use CTRL-D to reload.",
        )

    def _ensure_raw_propmt(self):
        # similar to _interrupt_to_raw_prompt, but assumes we are already in a prompt
        self._forward_unexpected_output()
        self._connection.write(RAW_MODE_CMD)
        prompt = self._connection.read_until(FIRST_RAW_PROMPT_SUFFIX, 1, True)
        if not prompt.endswith(FIRST_RAW_PROMPT_SUFFIX):
            raise TimeoutError("Could not ensure raw prompt")

    def _execute(self, script, capture_output=False):
        self._ensure_raw_propmt()

        # send command
        self._connection.write(script.encode(ENCODING) + EOT)
        debug("Wrote " + script + "\n--------\n")

        # fetch command confirmation
        ok = self._connection.read(2)
        debug("GOTOK")
        assert ok == OK, "Expected OK, got %r, followed by %r" % (ok, self._connection.read_all())
        return self._process_until_raw_prompt(capture_output)

    def _execute_without_output(self, script):
        out, err, value = self._execute(script, capture_output=True)
        if err or out:
            raise RuntimeError("Failed MP script: " + str(out) + "\n" + str(err))

        return value

    def _execute_print_expr(self, expr, prelude="", cleanup="", capture_output=False):
        # assuming expr really contains an expression
        # separator is for separating side-effect output and printed value
        script = ""
        if prelude:
            script += prelude + "\n"
        script += "print(%r, repr(%s), sep='', end=%r)" % (
            THONNY_MSG_START.decode(),
            expr,
            THONNY_MSG_END.decode(),
        )

        # assuming cleanup doesn't cause output
        if cleanup:
            script += "\n" + cleanup

        return self._execute(script, capture_output)

    def _evaluate(self, expr, prelude="", cleanup=""):
        _, _, value_repr = self._execute_print_expr(expr, prelude, cleanup)
        if value_repr is None:
            return None
        else:
            return ast.literal_eval(value_repr)

    def _process_until_initial_raw_prompt(self):
        self._connection.write(RAW_MODE_CMD)
        try:
            self._process_until_raw_prompt()
        except KeyboardInterrupt:
            self._interrupt()

    def _process_until_raw_prompt(self, capture_output=False):
        """
        Forwards output, extracts Thonny message, replaces normal prompts with raw prompts.
        
        This is executed when some code is running or just after requesting raw prompt.
        
        After submitting commands to the raw REPL, the output should be like
        {stdout}\x04\{stderr}\x04\n\>
        In the end of {stdout} there may be \x02{value-for-thonny}
        
        Interrupts will alter the execution, but from the response parsing
        perspective they don't matter as they look like any other exception.
        
        Things get complicated because of soft-reboots, which always end with
        regular prompt. Soft-reboots can occur because of Ctrl+D, machine.soft_reset()
        and even reset button (micro:bit).
        
        Because of soft-reboot we can't assume we'll find the terminating markers for 
        each command.
        
        Output produced by background threads (eg. in WiPy ESP32) cause even more difficulties, 
        because it becomes impossible to say whether we are at prompt and output
        is from another thread or the main thread is running.
        For now I'm ignoring these problems and assume all output comes from the main thread.
         
        """
        # TODO: experiment with Ctrl+C, Ctrl+D, reset
        eot_count = 0
        value = None
        done = False
        output = b""
        out = b""
        err = b""
        while not done:
            if (
                self._connection.num_bytes_received == 0
                and not self._ctrl_suggestion_given
                and time.time() - self._startup_time > 1.5
            ):
                self._show_error(
                    "\n"
                    + "Device is busy or does not respond. Your options:\n\n"
                    + "  - wait until it completes current work;\n"
                    + "  - use Ctrl+C to interrupt current work;\n"
                    + "  - use Stop/Restart to interrupt more and enter REPL.\n"
                )
                self._ctrl_suggestion_given = True

            # There may be an input submission waiting
            # and we can't progress without resolving it first
            self._check_for_side_commands()
            self._check_for_interrupt("device")

            # Process input in chunks (max 1 parsing marker per chunk).
            # Prefer whole lines (to reduce the number of events),
            # but don't wait too long for eol.
            ddd = self._connection.soft_read_until(BLOCK_CLOSERS, timeout=0.05)
            output += ddd
            stream_name = "stderr" if eot_count == 1 else "stdout"

            if output.endswith(THONNY_MSG_START):
                debug("MSGSTA: " + str(output))
                output = output[: -len(THONNY_MSG_START)]

                # Low chance of failure (eg. because of precisely timed reboot),
                # therefore it's safe to use big timeout
                temp = self._connection.soft_read_until(THONNY_MSG_END, timeout=3)
                if temp.endswith(THONNY_MSG_END):
                    value = temp[: -len(THONNY_MSG_END)]
                    debug("GOTVALUE: " + str(value))
                else:
                    # failure, restore everything to help diagnosis
                    output = output + THONNY_MSG_START + temp

            elif output.endswith(EOT):
                debug("EOT: " + str(output))
                output = output[: -len(EOT)]
                eot_count += 1
                if eot_count == 2:
                    # Normal completion of the command
                    # big chance of being at the raw prompt
                    temp = self._connection.soft_read_until(RAW_PROMPT, timeout=0.1)
                    if temp == RAW_PROMPT and self._connection.incoming_is_empty():
                        done = True
                    elif temp:
                        # Failure, temp needs to be parsed again
                        self._connection.unread(temp)

            elif output.endswith(FIRST_RAW_PROMPT) and self._connection.incoming_is_empty():
                debug("FIRAPRO: " + str(output))
                output = output[: -len(FIRST_RAW_PROMPT)]
                done = True

            elif (
                output.endswith(NORMAL_PROMPT)
                and self._connection.peek_incoming() == b"\r\n" + FIRST_RAW_PROMPT
            ):
                debug("NOPRO: " + str(output))
                output = output + self._connection.read_until(FIRST_RAW_PROMPT)
                # skip both normal and raw prompt together
                # (otherwise they get processed separately)
                output = output[: -len(NORMAL_PROMPT + b"\r\n" + FIRST_RAW_PROMPT)]
                done = True

            elif output.endswith(NORMAL_PROMPT) and self._connection.incoming_is_empty():
                debug("NOPRO2: " + str(output))
                output = output[: -len(NORMAL_PROMPT)]
                # switch to raw mode and continue
                self._connection.write(RAW_MODE_CMD)

            if output.endswith(FIRST_RAW_PROMPT[:-1]):
                # incomplete raw prompt, wait for more
                pass
            elif output:
                if capture_output:
                    if stream_name == "stdout":
                        out += output
                    else:
                        assert stream_name == "stderr"
                        err += output
                else:
                    self._send_output(output.decode(ENCODING, errors="replace"), stream_name)
                output = b""

        debug("doneproc")
        return (
            out.decode(ENCODING, errors="replace"),
            err.decode(ENCODING, errors="replace"),
            None if value is None else value.decode(ENCODING),
        )

    def _clear_environment(self):
        # TODO: Ctrl+D in raw repl is perfect for MicroPython
        # but on CircuitPython it runs main.py

        # TODO: which is better:
        # self._execute_async(dedent("""
        #    for name in globals():
        #        if not name.startswith("__"):
        #            del globals()[name]
        # """).strip())
        # or
        self._execute("globals().clear(); __name__ = '__main__'")

    def _check_for_side_commands(self):
        # most likely the queue is empty
        if self._command_queue.empty():
            return

        postponed = []
        while not self._command_queue.empty():
            cmd = self._command_queue.get()
            if isinstance(cmd, InputSubmission):
                self._submit_input(cmd.data)
            elif isinstance(cmd, EOFCommand):
                self._soft_reboot(True)
            else:
                postponed.append(cmd)

        # put back postponed commands
        while postponed:
            self._command_queue.put(postponed.pop(0))

    def _forward_unexpected_output(self):
        "Invoked between commands"
        data = self._connection.read_all()
        if data.endswith(NORMAL_PROMPT):
            # looks like the device was resetted

            # hide the regular prompt from the output ...
            data = data[: -len(NORMAL_PROMPT)]
            at_prompt = True
        else:
            at_prompt = False

        self._send_output(data.decode(ENCODING, "replace"), "stdout")
        if at_prompt:
            # ... and recreate Thonny prompt
            self.send_message(ToplevelResponse())

        self._check_for_connection_errors()

    def _supports_directories(self):
        # NB! make sure self._cwd is queried first
        return bool(self._cwd)

    def _connected_to_microbit(self):
        return "micro:bit" in self._welcome_text.lower()

    def _cmd_interrupt(self, cmd):
        self._interrupt()

    def _cmd_cd(self, cmd):
        if len(cmd.args) == 1:
            if not self._supports_directories():
                raise UserError("This device doesn't have directories")

            path = cmd.args[0]
            self._execute("import os as __thonny_os; __thonny_os.chdir(%r)" % path)
            self._cwd = self._fetch_cwd()
            return {}
        else:
            raise UserError("%cd takes one parameter")

    def _cmd_Run(self, cmd):
        self._clear_environment()
        assert cmd.get("source")
        self._execute(cmd["source"])
        return {}

    def _cmd_execute_source(self, cmd):
        try:
            # Try to parse as expression
            ast.parse(cmd.source, mode="eval")
            # If it didn't fail then source is an expression
            _, _, value_repr = self._execute_print_expr(cmd.source)
            if value_repr is None:
                value_repr = repr(None)
            return {"value_info": ValueInfo(0, value_repr)}
        except SyntaxError:
            # source is a statement (or invalid syntax)
            self._execute(cmd.source)
            return {}

    def _cmd_execute_system_command(self, cmd):
        # Can't use stdin, because a thread is draining it
        execute_system_command(cmd, cwd=self._local_cwd, disconnect_stdin=True)

    def _cmd_get_globals(self, cmd):
        if cmd.module_name == "__main__":
            globs = self._evaluate(
                "{name : repr(value) for (name, value) in globals().items() if not name.startswith('__')}"
            )
        else:
            globs = self._evaluate(
                "{name : repr(getattr(__mod_for_globs, name)) in dir(__mod_for_globs) if not name.startswith('__')}",
                prelude="import %s as __mod_for_globs",
            )
        return {"module_name": cmd.module_name, "globals": globs}

    def _cmd_get_dirs_child_data(self, cmd):
        if self._supports_directories():
            data = self._get_dirs_child_data_generic(cmd["paths"])
            dir_separator = "/"
        else:
            assert cmd["paths"] == {""}, "Bad command: " + repr(cmd)
            sizes = self._get_microbit_file_sizes()
            root_data = {name: {"kind": "file", "size": size} for (name, size) in sizes.items()}
            data = {"": root_data}
            dir_separator = ""

        return {"node_id": cmd["node_id"], "dir_separator": dir_separator, "data": data}

    def _cmd_get_fs_info(self, cmd):
        return self._get_fs_info(cmd.path)

    def _cmd_write_file(self, cmd):
        def generate_blocks(content_bytes, block_size):
            for i in range(0, len(content_bytes), block_size):
                yield content_bytes[i : i + block_size]

        self._write_file(generate_blocks(cmd["content_bytes"], BUFFER_SIZE), cmd["path"])

        return InlineResponse(
            command_name="write_file", path=cmd["path"], editor_id=cmd.get("editor_id")
        )

    def _cmd_delete(self, cmd):
        assert cmd.paths

        paths = sorted(cmd.paths, key=lambda x: len(x), reverse=True)

        try:
            self._delete_via_serial(paths)
        except Exception as e:
            if "read-only" in str(e).lower():
                self._delete_via_mount(paths)

        self._sync_all_filesystems()

    def _internal_path_to_mounted_path(self, path):
        mount_path = self._get_fs_mount()
        if mount_path is None:
            return None

        flash_prefix = self._get_flash_prefix()
        if not path.startswith(flash_prefix):
            return None

        path_suffix = path[len(flash_prefix) :]

        return os.path.join(mount_path, os.path.normpath(path_suffix))

    def _cmd_read_file(self, cmd):
        try:
            content_bytes = b"".join(self._read_file(cmd["path"]))
            error = None
        except Exception as e:
            _report_internal_error()
            error = str(e)
            content_bytes = None

        return {"content_bytes": content_bytes, "path": cmd["path"], "error": error}

    def _cmd_download(self, cmd):
        total_size = 0
        completed_files_size = 0
        remote_files = self._list_remote_files_with_info(cmd["source_paths"])
        target_dir = cmd["target_dir"].rstrip("/").rstrip("\\")

        download_items = []
        for file in remote_files:
            total_size += file["size"]
            # compute filenames (and subdirs) in target_dir
            # relative to the context of the user selected items
            assert file["path"].startswith(file["original_context"])
            path_suffix = file["path"][len(file["original_context"]) :].strip("/").strip("\\")
            target_path = os.path.join(target_dir, os.path.normpath(path_suffix))
            download_items.append(dict(source=file["path"], target=target_path, size=file["size"]))

        if not cmd["allow_overwrite"]:
            targets = [item["target"] for item in download_items]
            existing_files = list(filter(os.path.exists, targets))
            if existing_files:
                return {
                    "existing_files": existing_files,
                    "source_paths": cmd["source_paths"],
                    "target_dir": cmd["target_dir"],
                    "description": cmd["description"],
                }

        def notify(current_file_progress):
            self._check_send_inline_progress(
                cmd, completed_files_size + current_file_progress, total_size
            )

        # replace the indeterminate progressbar with determinate as soon as possible
        notify(0)

        for item in download_items:
            written_bytes = self._download_file(item["source"], item["target"], notify)
            assert written_bytes == item["size"]
            completed_files_size += item["size"]

    def _cmd_upload(self, cmd):
        completed_files_size = 0
        local_files = self._list_local_files_with_info(cmd["source_paths"])
        target_dir = cmd["target_dir"]
        assert target_dir.startswith("/") or not self._supports_directories()
        assert not target_dir.endswith("/") or target_dir == "/"

        upload_items = []
        for file in local_files:
            # compute filenames (and subdirs) in target_dir
            # relative to the context of the user selected items
            assert file["path"].startswith(file["original_context"])
            path_suffix = file["path"][len(file["original_context"]) :].strip("/").strip("\\")
            target_path = self._join_remote_path_parts(target_dir, to_remote_path(path_suffix))
            upload_items.append(dict(source=file["path"], target=target_path, size=file["size"]))

        if not cmd["allow_overwrite"]:
            targets = [item["target"] for item in upload_items]
            existing_files = self._get_existing_remote_files(targets)
            if existing_files:
                return {
                    "existing_files": existing_files,
                    "source_paths": cmd["source_paths"],
                    "target_dir": cmd["target_dir"],
                    "description": cmd["description"],
                }

        total_size = sum([item["size"] for item in upload_items])

        def notify(current_file_progress):
            self._check_send_inline_progress(
                cmd, completed_files_size + current_file_progress, total_size
            )

        # replace the indeterminate progressbar with determinate as soon as possible
        notify(0)

        for item in upload_items:
            written_bytes = self._upload_file(item["source"], item["target"], notify)
            assert written_bytes == item["size"]
            completed_files_size += item["size"]

    def _cmd_mkdir(self, cmd):
        assert self._supports_directories()
        assert cmd.path.startswith("/")
        self._makedirs(cmd.path)
        self._sync_all_filesystems()

    def _cmd_editor_autocomplete(self, cmd):
        # template for the response
        result = dict(source=cmd.source, row=cmd.row, column=cmd.column)

        try:
            script = jedi.Script(cmd.source, cmd.row, cmd.column, sys_path=[self._api_stubs_path])
            completions = script.completions()
            result["completions"] = self._filter_completions(completions)
        except Exception:
            traceback.print_exc()
            result["error"] = "Autocomplete error"

        return result

    def _filter_completions(self, completions):
        # filter out completions not applicable to MicroPython
        result = []
        for completion in completions:
            if completion.name.startswith("__"):
                continue

            if completion.parent() and completion.full_name:
                parent_name = completion.parent().name
                name = completion.name
                root = completion.full_name.split(".")[0]

                # jedi proposes names from CPython builtins
                if root in self._builtins_info and name not in self._builtins_info[root]:
                    continue

                if parent_name == "builtins" and name not in self._builtins_info:
                    continue

            result.append({"name": completion.name, "complete": completion.complete})

        return result

    def _cmd_shell_autocomplete(self, cmd):
        source = cmd.source

        # TODO: combine dynamic results and jedi results
        if source.strip().startswith("import ") or source.strip().startswith("from "):
            # this needs the power of jedi
            response = {"source": cmd.source}

            try:
                # at the moment I'm assuming source is the code before cursor, not whole input
                lines = source.split("\n")
                script = jedi.Script(
                    source, len(lines), len(lines[-1]), sys_path=[self._api_stubs_path]
                )
                completions = script.completions()
                response["completions"] = self._filter_completions(completions)
            except Exception:
                traceback.print_exc()
                response["error"] = "Autocomplete error"

            return response
        else:
            # use live data
            match = re.search(
                r"(\w+\.)*(\w+)?$", source
            )  # https://github.com/takluyver/ubit_kernel/blob/master/ubit_kernel/kernel.py
            if match:
                prefix = match.group()
                if "." in prefix:
                    obj, prefix = prefix.rsplit(".", 1)
                    names = self._evaluate(
                        "dir({}) if '{}' in locals() or '{}' in globals() else []".format(
                            obj, obj, obj
                        )
                    )
                else:
                    names = self._evaluate("dir()")
            else:
                names = []
                prefix = ""

            completions = []

            # prevent TypeError (iterating over None)
            names = names if names else []

            for name in names:
                if name.startswith(prefix) and not name.startswith("__"):
                    completions.append({"name": name, "complete": name[len(prefix) :]})

            return {"completions": completions, "source": source}

    def _cmd_dump_api_info(self, cmd):
        "For use during development of the plug-in"

        self._execute(
            dedent(
                """
            def __get_object_atts(obj):
                result = []
                errors = []
                for name in dir(obj):
                    try:
                        val = getattr(obj, name)
                        result.append((name, repr(val), repr(type(val))))
                    except BaseException as e:
                        errors.append("Couldn't get attr '%s' from object '%r', Err: %r" % (name, obj, e))
                return (result, errors)
        """
            )
        )

        for module_name in sorted(self._fetch_builtin_modules()):
            if (
                not module_name.startswith("_")
                and not module_name.startswith("adafruit")
                # and not module_name == "builtins"
            ):
                file_name = os.path.join(
                    self._api_stubs_path, module_name.replace(".", "/") + ".py"
                )
                self._dump_module_stubs(module_name, file_name)

    def _dump_module_stubs(self, module_name, file_name):
        out, err, __ = self._execute("import {0}".format(module_name), capture_output=True)
        if out or err:
            print("FAILED IMPORTING MODULE:", module_name, "\nErr: " + out + err)
            return

        os.makedirs(os.path.dirname(file_name), exist_ok=True)
        with io.open(file_name, "w", encoding="utf-8", newline="\n") as fp:
            if module_name not in [
                "webrepl",
                "_webrepl",
                "gc",
                "http_client",
                "http_client_ssl",
                "http_server",
                "framebuf",
                "example_pub_button",
                "flashbdev",
            ]:
                self._dump_object_stubs(fp, module_name, "")

    def _dump_object_stubs(self, fp, object_expr, indent):
        if object_expr in [
            "docs.conf",
            "pulseio.PWMOut",
            "adafruit_hid",
            "upysh",
            # "webrepl",
            # "gc",
            # "http_client",
            # "http_server",
        ]:
            print("SKIPPING problematic name:", object_expr)
            return

        print("DUMPING", indent, object_expr)
        items, errors = self._evaluate("__get_object_atts({0})".format(object_expr))

        if errors:
            print("ERRORS", errors)

        for name, rep, typ in sorted(items, key=lambda x: x[0]):
            if name.startswith("__"):
                continue

            print("DUMPING", indent, object_expr, name)
            self._send_text_to_shell("  * " + name + " : " + typ, "stdout")

            if typ in ["<class 'function'>", "<class 'bound_method'>"]:
                fp.write(indent + "def " + name + "():\n")
                fp.write(indent + "    pass\n\n")
            elif typ in ["<class 'str'>", "<class 'int'>", "<class 'float'>"]:
                fp.write(indent + name + " = " + rep + "\n")
            elif typ == "<class 'type'>" and indent == "":
                # full expansion only on toplevel
                fp.write("\n")
                fp.write(indent + "class " + name + ":\n")  # What about superclass?
                fp.write(indent + "    ''\n")
                self._dump_object_stubs(fp, "{0}.{1}".format(object_expr, name), indent + "    ")
            else:
                # keep only the name
                fp.write(indent + name + " = None\n")

    def _read_file(self, path):
        # TODO: read from mount when possible
        # file_size = self._get_file_size(path)
        block_size = 512

        self._execute_without_output("__thonny_fp = open(%r, 'rb')" % path)
        if "binascii" in self._builtin_modules:
            self._execute_without_output("from binascii import hexlify as __temp_hexlify")

        while True:
            self._check_for_interrupt("local")
            if "binascii" in self._builtin_modules:
                block = binascii.unhexlify(
                    self._evaluate("__temp_hexlify(__thonny_fp.read(%s))" % block_size)
                )
            else:
                block = self._evaluate("__thonny_fp.read(%s)" % block_size)
            if block:
                yield block
            if len(block) < block_size:
                break

        self._execute_without_output(
            dedent(
                """
            __thonny_fp.close()
            del __thonny_fp
            try:
                del __temp_hexlify
            except:
                pass
            """
            )
        )

    def _write_file(self, content_blocks, target_path, notifier=None):
        try:
            result = self._write_file_via_serial(content_blocks, target_path, notifier)
        except ReadOnlyFilesystemError:
            result = self._write_file_via_mount(content_blocks, target_path, notifier)

        self._sync_all_filesystems()
        return result

    def _write_file_via_mount(self, content_blocks, target_path, notifier=None):
        mounted_target_path = self._internal_path_to_mounted_path(target_path)
        with open(mounted_target_path, "wb") as f:
            bytes_written = 0
            for block in content_blocks:
                self._check_for_interrupt("local")
                bytes_written += f.write(block)
                f.flush()
                os.fsync(f)
                if notifier is not None:
                    notifier(bytes_written)

        return bytes_written

    def _write_file_via_serial(self, content_blocks, target_path, notifier=None):
        # prelude
        try:
            _, err, _ = self._execute(
                dedent(
                    """
                __thonny_path = '{path}'
                __thonny_written = 0
                __thonny_fp = open(__thonny_path, 'wb')
                """
                ).format(path=target_path),
                capture_output=True,
            )

            if "readonly" in err.replace("-", "").lower():
                raise ReadOnlyFilesystemError()

            elif err:
                raise RuntimeError("Problem opening file for writing: " + err)

            # Define function to allow shorter write commands
            if "binascii" in self._builtin_modules:
                self._execute_without_output(
                    dedent(
                        """
                    from binascii import unhexlify as __thonny_unhex
                    def __W(x):
                        global __thonny_written
                        __thonny_written += __thonny_fp.write(__thonny_unhex(x))
                        __thonny_fp.flush()
                """
                    )
                )
            else:
                self._execute_without_output(
                    dedent(
                        """
                    def __W(x):
                        global __thonny_written
                        __thonny_written += __thonny_fp.write(x)
                """
                    )
                )

            bytes_sent = 0
            for block in content_blocks:
                self._check_for_interrupt("local")
                if "binascii" in self._builtin_modules:
                    script = "__W(%r)" % binascii.hexlify(block)
                else:
                    script = "__W(%r)" % block
                self._execute_without_output(script)
                bytes_sent += len(block)
                if notifier is not None:
                    notifier(bytes_sent)

            bytes_received = self._evaluate("__thonny_written")

            if bytes_received != bytes_sent:
                raise UserError(
                    "Expected %d written bytes but wrote %d" % (bytes_sent, bytes_received)
                )

        finally:
            # clean up
            self._execute(
                dedent(
                    """
                    try:
                        del __W
                        del __thonny_written
                        del __thonny_path
                        __thonny_fp.close()
                        del __thonny_fp
                        del __thonny_unhex
                    except:
                        pass
                """
                )
            )

        return bytes_sent

    def _sync_all_filesystems(self):
        self._execute_without_output(
            dedent(
                """
            try:
                from os import sync as __thonny_sync
                __thonny_sync()
                del __thonny_sync
            except ImportError:
                pass
        """
            )
        )

    def _list_local_files_with_info(self, paths):
        def rec_list_with_size(path):
            result = {}
            if os.path.isfile(path):
                result[path] = os.path.getsize(path)
            elif os.path.isdir(path):
                for name in os.listdir(path):
                    result.update(rec_list_with_size(os.path.join(path, name)))
            else:
                raise RuntimeError("Can't process " + path)

            return result

        result = []
        for requested_path in paths:
            sizes = rec_list_with_size(requested_path)
            for path in sizes:
                result.append(
                    {
                        "path": path,
                        "size": sizes[path],
                        "original_context": os.path.dirname(requested_path),
                    }
                )

        result.sort(key=lambda rec: rec["path"])
        return result

    def _list_remote_files_with_info(self, paths):
        # prepare universal functions
        self._execute_without_output(
            dedent(
                """
            try:
                import os as __thonny_os
                from os import stat as __thonny_stat
                
                def __thonny_getsize(path):
                    return __thonny_stat(path)[6]
                
                def __thonny_isdir(path):
                    return __thonny_stat(path)[0] & 0o170000 == 0o040000
                    
            except ImportError:
                __thonny_stat = None
                # micro:bit
                from os import size as __thonny_getsize
                
                def __thonny_isdir(path):
                    return False
        """
            )
        )

        self._execute_without_output(
            dedent(
                """
            def __thonny_rec_list_with_size(path):
                result = {}
                if __thonny_isdir(path):
                    for name in __thonny_os.listdir(path):
                        result.update(__thonny_rec_list_with_size(path + "/" + name))
                else:
                    result[path] = __thonny_getsize(path)
    
                return result
        """
            )
        )

        result = []
        for requested_path in paths:
            sizes = self._evaluate("__thonny_rec_list_with_size(%r)" % requested_path)
            for path in sizes:
                result.append(
                    {
                        "path": path,
                        "size": sizes[path],
                        "original_context": os.path.dirname(requested_path),
                    }
                )

        result.sort(key=lambda rec: rec["path"])

        self._execute_without_output(
            dedent(
                """
                del __thonny_os
                del __thonny_stat
                del __thonny_getsize
                del __thonny_isdir
                del __thonny_rec_list_with_size
            """
            )
        )
        return result

    def _get_existing_remote_files(self, paths):
        if self._supports_directories():
            func = "stat"
        else:
            func = "size"

        return self._evaluate(
            "__thonny_result",
            prelude=dedent(
                """
                import os as __thonny_os
                __thonny_result = []
                for __thonny_path in %r:
                    try:
                        __thonny_os.%s(__thonny_path)
                        __thonny_result.append(__thonny_path)
                    except OSError:
                        pass
                """
            )
            % (paths, func),
            cleanup=dedent(
                """
                del __thonny_os
                del __thonny_result
                del __thonny_path
                """
            ),
        )

    def _join_remote_path_parts(self, left, right):
        if left == "":  # micro:bit
            assert not self._supports_directories()
            return right.strip("/")

        return left.rstrip("/") + "/" + right.strip("/")

    def _get_file_size(self, path):
        if self._supports_directories():
            script = "__thonny_os.stat(%r)[6]"
        else:
            script = "os.stat(%r)[6]"

        return self._evaluate(script % path, prelude="import os as __thonny_os")

    def _makedirs(self, path):
        if path == "/":
            return

        try:
            self._makedirs_via_serial(path)
        except Exception as e:
            if "read-only" in str(e).lower():
                self._makedirs_via_mount(path)

    def _makedirs_via_mount(self, path):
        mounted_path = self._internal_path_to_mounted_path(path)
        assert mounted_path is not None, "Couldn't find mounted path for " + path
        os.makedirs(mounted_path, exist_ok=True)

    def _makedirs_via_serial(self, path):
        if path == "/":
            return
        path = path.rstrip("/")

        script = (
            dedent(
                """
            import os as __thonny_os
            __thonny_parts = %r.split('/')
            for i in range(2, len(__thonny_parts) + 1):
                __thonny_path = "/".join(__thonny_parts[:i])
                try:
                    __thonny_os.stat(__thonny_path)
                except OSError:
                    # does not exist
                    __thonny_os.mkdir(__thonny_path)
            
            del __thonny_parts
            try:
                del __thonny_path
            except:
                pass
        """
            )
            % path
        )

        self._execute_without_output(script)

    def _delete_via_mount(self, paths):
        for path in paths:
            mounted_path = self._internal_path_to_mounted_path(path)
            assert mounted_path is not None
            shutil.rmtree(mounted_path)

    def _delete_via_serial(self, paths):
        if not self._supports_directories():
            self._execute_without_output(
                dedent(
                    """
                import os as __thonny_os
                for __thonny_path in %r: 
                    __thonny_os.remove(__thonny_path)
                    
                del __thonny_path
                del __thonny_os
            """
                )
                % paths
            )
        else:
            self._execute_without_output(
                dedent(
                    """
                import os as __thonny_os
                def __thonny_delete(path):
                    if __thonny_os.stat(path)[0] & 0o170000 == 0o040000:
                        for name in __thonny_os.listdir(path):
                            child_path = path + "/" + name
                            __thonny_delete(child_path)
                        __thonny_os.rmdir(path)
                    else:
                        __thonny_os.remove(path)
                
                for __thonny_path in %r: 
                    __thonny_delete(__thonny_path)
                    
                del __thonny_path
                del __thonny_delete
                del __thonny_os
            """
                )
                % paths
            )

    def _upload_file(self, source, target, notifier):
        assert target.startswith("/") or not self._supports_directories()
        target_dir, _ = linux_dirname_basename(target)
        assert target_dir.startswith("/") or not self._supports_directories()

        self._makedirs(target_dir)

        def block_generator():
            with open(source, "rb") as source_fp:
                while True:
                    block = source_fp.read(512)
                    if block:
                        yield block
                    else:
                        break

        return self._write_file(block_generator(), target, notifier=notifier)

    def _download_file(self, source, target, notifier=None):
        os.makedirs(os.path.dirname(target), exist_ok=True)
        bytes_written = 0
        with open(target, "wb") as out_fp:
            for block in self._read_file(source):
                out_fp.write(block)
                os.fsync(out_fp)
                bytes_written += len(block)
                notifier(bytes_written)

        return bytes_written

    def _get_fs_mount_label(self):
        # This method is most likely required with CircuitPython,
        # so try its approach first
        # https://learn.adafruit.com/welcome-to-circuitpython/the-circuitpy-drive
        result = self._evaluate(
            "__thonny_result",
            prelude=dedent(
                """
            try:
                from storage import getmount as __thonny_getmount
                try:
                    __thonny_result = __thonny_getmount("/").label
                finally:
                    del __thonny_getmount
            except ImportError:
                __thonny_result = None 
            except OSError:
                __thonny_result = None 
        """
            ),
            cleanup="del __thonny_result",
        )

        if result is not None:
            return result

        if self._welcome_text is None:
            return None

        """
        # following is not reliable and probably not needed 
        markers_by_name = {"PYBFLASH": {"pyb"}, "CIRCUITPY": {"circuitpython"}}

        for name in markers_by_name:
            for marker in markers_by_name[name]:
                if marker.lower() in self._welcome_text.lower():
                    return name
        """

        return None

    def _get_flash_prefix(self):
        if not self._supports_directories():
            return ""
        elif (
            "LoBo" in self._welcome_text
            or "WiPy with ESP32" in self._welcome_text
            or "PYBLITE" in self._welcome_text
            or "PYBv" in self._welcome_text
            or "PYBOARD" in self._welcome_text.upper()
        ):
            return "/flash/"
        else:
            return "/"

    def _get_fs_mount(self):
        label = self._get_fs_mount_label()
        if label is None:
            return None
        else:
            candidates = find_volumes_by_name(
                self._get_fs_mount_label(),
                # querying A can be very slow
                skip_letters="A",
            )
            if len(candidates) == 0:
                raise RuntimeError("Could not find volume " + self._get_fs_mount_label())
            elif len(candidates) > 1:
                raise RuntimeError("Found several possible mount points: %s" % candidates)
            else:
                return candidates[0]

    def _get_fs_info(self, path):
        result = self._evaluate(
            dedent(
                """{
                    "total" : __thonny_total,
                    "used" : __thonny_used,
                    "free": __thonny_free,
                    "sizes": __thonny_sizes
                }"""
            ),
            prelude=dedent(
                """
                try:
                    from os import statvfs as __thonny_statvfs
                    __thonny_stat = __thonny_statvfs(%r)
                    __thonny_total = __thonny_stat[2] * __thonny_stat[0]
                    __thonny_free = __thonny_stat[3] * __thonny_stat[0]
                    __thonny_used = __thonny_total - __thonny_free
                    __thonny_sizes = None
                    del __thonny_statvfs
                    del __thonny_stat 
                except ImportError:
                    import os as __thonny_os
                    __thonny_sizes = [__thonny_os.size(name) for name in __thonny_os.listdir()]
                    __thonny_used = None
                    __thonny_total = None
                    __thonny_free = None  
                    del __thonny_os
            """
            )
            % path,
            cleanup=dedent(
                """
                del __thonny_total
                del __thonny_free
                del __thonny_used
                del __thonny_sizes
            """
            ),
        )

        if result["sizes"] is not None:
            if self._connected_to_microbit():
                comment = "Assuming around 30 kB of storage space for user files."
            else:
                comment = "Don't know the size of storage space on this device."

            files_total_size = sum(result["sizes"])

            # TODO: compute number of used blocks
            if files_total_size > 0:
                comment += "\n\n" + "At least %s of it is used by %d file(s)." % (
                    sizeof_fmt(files_total_size),
                    len(result["sizes"]),
                )

            result["comment"] = comment
            del result["sizes"]

        return result

    def _get_microbit_file_sizes(self):
        return self._evaluate(
            "{name : __thonny_os.size(name) for name in __thonny_os.listdir()}",
            prelude="import os as __thonny_os",
            cleanup="del __thonny_os",
        )

    def _get_dirs_child_data_generic(self, paths):
        return self._evaluate(
            "__thonny_result",
            prelude=dedent(
                """
                import os as __thonny_os
                # Init all vars, so that they can be deleted
                # even if the loop makes no iterations
                __thonny_result = {}
                __thonny_path = None
                __thonny_st = None 
                __thonny_child_names = None
                __thonny_children = None
                __thonny_name = None
                __thonny_real_path = None
                __thonny_full = None
                
                for __thonny_path in %(paths)r:
                    __thonny_real_path = __thonny_path or '/'
                    try:
                        __thonny_child_names = __thonny_os.listdir(__thonny_real_path)
                    except OSError:
                        # probably deleted directory
                        __thonny_children = None
                    else:
                        __thonny_children = {}
                        for __thonny_name in __thonny_child_names:
                            if __thonny_name.startswith('.') or __thonny_name == "System Volume Information":
                                continue
                            __thonny_full = (__thonny_real_path + '/' + __thonny_name).replace("//", "/")
                            try:
                                __thonny_st = __thonny_os.stat(__thonny_full)
                                if __thonny_st[0] & 0o170000 == 0o040000:
                                    # directory
                                    __thonny_children[__thonny_name] = {"kind" : "dir", "size" : None}
                                else:
                                    __thonny_children[__thonny_name] = {"kind" : "file", "size" :__thonny_st[6]}
                                
                                # converting from 2000-01-01 epoch to Unix epoch 
                                __thonny_children[__thonny_name]["time"] = max(__thonny_st[8], __thonny_st[9]) + 946684800
                            except OverflowError:
                                # Probably "System Volume Information" in trinket
                                # https://github.com/thonny/thonny/issues/923
                                pass
                            
                    __thonny_result[__thonny_path] = __thonny_children                            
            """
            )
            % {"paths": paths},
            cleanup=dedent(
                """
                del __thonny_os
                del __thonny_st
                del __thonny_children
                del __thonny_name
                del __thonny_path
                del __thonny_full
                del __thonny_result
                del __thonny_real_path
            """
            ),
        )

    def _check_for_connection_errors(self):
        self._connection._check_for_error()

    def _on_connection_closed(self, error=None):
        message = "Connection lost"
        if error:
            message += " (" + str(error) + ")"
        self._send_output("\n" + message + "\n", "stderr")
        self._send_output("\n" + "Use Stop/Restart to reconnect." + "\n", "stderr")
        sys.exit(EXPECTED_TERMINATION_CODE)

    def _show_error(self, msg):
        self._send_output(msg + "\n", "stderr")


class ExecutionError(Exception):
    pass


def _report_internal_error():
    print("PROBLEM WITH THONNY'S BACK-END:\n", file=sys.stderr)
    traceback.print_exc()


def parse_api_information(file_path):
    with tokenize.open(file_path) as fp:
        source = fp.read()

    tree = ast.parse(source)

    defs = {}

    # TODO: read also docstrings ?

    for toplevel_item in tree.body:
        if isinstance(toplevel_item, ast.ClassDef):
            class_name = toplevel_item.name
            member_names = []
            for item in toplevel_item.body:
                if isinstance(item, ast.FunctionDef):
                    member_names.append(item.name)
                elif isinstance(item, ast.Assign):
                    # TODO: check Python 3.4
                    "TODO: item.targets[0].id"

            defs[class_name] = member_names

    return defs


def linux_dirname_basename(path):
    if path == "/":
        return ("/", "")

    if "/" not in path:  # micro:bit
        return "", path

    path = path.rstrip("/")
    dir_, file_ = path.rsplit("/", maxsplit=1)
    if dir_ == "":
        dir_ = "/"

    return dir_, file_


def to_remote_path(path):
    return path.replace("\\", "/")


class ReadOnlyFilesystemError(RuntimeError):
    pass


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--clean", type=lambda s: True if s == "True" else False)
    parser.add_argument("--port", type=str)
    parser.add_argument("--url", type=str)
    parser.add_argument("--password", type=str)
    parser.add_argument("--api_stubs_path", type=str)
    args = parser.parse_args()

    port = None if args.port == "None" else args.port
    try:
        if port is None:
            # remain busy
            while True:
                time.sleep(1000)
        elif port == "webrepl":
            from thonny.plugins.micropython.webrepl_connection import WebReplConnection

            connection = WebReplConnection(args.url, args.password)
        else:
            from thonny.plugins.micropython.serial_connection import SerialConnection

            connection = SerialConnection(port, BAUDRATE)

        vm = MicroPythonBackend(connection, clean=args.clean, api_stubs_path=args.api_stubs_path)

    except ConnectionFailedException as e:
        text = "\n" + str(e) + "\n"
        msg = BackendEvent(event_type="ProgramOutput", stream_name="stderr", data=text)
        sys.stdout.write(serialize_message(msg) + "\n")
        sys.stdout.flush()



import ast
import logging
import tkinter as tk
from tkinter import ttk

import thonny.memory
from thonny import get_runner, get_workbench, ui_utils
from thonny.common import InlineCommand
from thonny.misc_utils import shorten_repr
from thonny.tktextext import TextFrame
import base64


class ObjectInspector(ttk.Frame):
    def __init__(self, master):
        ttk.Frame.__init__(self, master, style="ViewBody.TFrame")

        self.object_id = None
        self.object_info = None

        # self._create_general_page()
        self._create_content_page()
        self._create_attributes_page()
        self.active_page = self.content_page
        self.active_page.grid(row=1, column=0, sticky="nsew")

        toolbar = self._create_toolbar()
        toolbar.grid(row=0, column=0, sticky="nsew", pady=(0, 1))

        self.columnconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)

        get_workbench().bind("ObjectSelect", self.show_object, True)
        get_workbench().bind("get_object_info_response", self._handle_object_info_event, True)
        get_workbench().bind("DebuggerResponse", self._handle_progress_event, True)
        get_workbench().bind("ToplevelResponse", self._handle_progress_event, True)

        # self.demo()

    def _create_toolbar(self):
        toolbar = ttk.Frame(self, style="ViewToolbar.TFrame")

        self.title_label = ttk.Label(
            toolbar,
            style="ViewToolbar.TLabel",
            text=""
            # borderwidth=1,
            # background=ui_utils.get_main_background()
        )
        self.title_label.grid(row=0, column=3, sticky="nsew", pady=5, padx=5)
        toolbar.columnconfigure(3, weight=1)

        self.tabs = []

        def create_tab(col, caption, page):
            if page == self.active_page:
                style = "Active.ViewTab.TLabel"
            else:
                style = "Inactive.ViewTab.TLabel"
            tab = ttk.Label(toolbar, text=caption, style=style)
            tab.grid(row=0, column=col, pady=5, padx=5, sticky="nsew")
            self.tabs.append(tab)
            page.tab = tab

            def on_click(event):
                if self.active_page == page:
                    return
                else:
                    if self.active_page is not None:
                        self.active_page.grid_forget()
                        self.active_page.tab.configure(style="Inactive.ViewTab.TLabel")

                    self.active_page = page
                    page.grid(row=1, column=0, sticky="nsew", padx=0)
                    tab.configure(style="Active.ViewTab.TLabel")
                    if self.active_page == self.attributes_page and (
                        self.object_info is None or not self.object_info.get("attributes")
                    ):
                        self.request_object_info()

            tab.bind("<1>", on_click)

        # create_tab(1, "Overview", self.general_page)
        create_tab(5, "Data", self.content_page)
        create_tab(6, "Atts", self.attributes_page)

        def create_navigation_link(col, image_filename, action, tooltip, padx=0):
            button = ttk.Button(
                toolbar,
                # command=handler,
                image=get_workbench().get_image(image_filename),
                style="ViewToolbar.Toolbutton",  # TODO: does this cause problems in some Macs?
                state=tk.NORMAL,
            )
            ui_utils.create_tooltip(button, tooltip)

            button.grid(row=0, column=col, sticky=tk.NE, padx=padx, pady=4)
            button.bind("<Button-1>", action)
            return button

        def configure(event):
            if event.width > 20:
                self.title_label.configure(wraplength=event.width - 10)

        self.title_label.bind("<Configure>", configure, True)

        self.back_button = create_navigation_link(
            1, "nav-backward", self.navigate_back, _("Previous object"), (5, 0)
        )
        self.forward_button = create_navigation_link(
            2, "nav-forward", self.navigate_forward, _("Next object")
        )
        self.back_links = []
        self.forward_links = []

        return toolbar

    def _create_content_page(self):
        self.content_page = ttk.Frame(self, style="ViewBody.TFrame")
        # type-specific inspectors
        self.current_content_inspector = None
        self.content_inspectors = []
        # load custom inspectors
        for insp_class in get_workbench().content_inspector_classes:
            self.content_inspectors.append(insp_class(self.content_page))

        # read standard inspectors
        self.content_inspectors.extend(
            [
                FileHandleInspector(self.content_page),
                FunctionInspector(self.content_page),
                StringInspector(self.content_page),
                ElementsInspector(self.content_page),
                DictInspector(self.content_page),
                ImageInspector(self.content_page),
                ReprInspector(self.content_page),  # fallback content inspector
            ]
        )

        self.content_page.columnconfigure(0, weight=1)
        self.content_page.rowconfigure(0, weight=1)

    def _create_attributes_page(self):
        self.attributes_page = AttributesFrame(self)

    def navigate_back(self, event):
        if len(self.back_links) == 0:
            return

        self.forward_links.append(self.object_id)
        self._show_object_by_id(self.back_links.pop(), True)

    def navigate_forward(self, event):
        if len(self.forward_links) == 0:
            return

        self.back_links.append(self.object_id)
        self._show_object_by_id(self.forward_links.pop(), True)

    def show_object(self, event):
        self._show_object_by_id(event.object_id)

    def _show_object_by_id(self, object_id, via_navigation=False):

        if self.winfo_ismapped() and self.object_id != object_id:
            if not via_navigation and self.object_id is not None:
                if self.object_id in self.back_links:
                    self.back_links.remove(self.object_id)
                self.back_links.append(self.object_id)
                del self.forward_links[:]

            self.object_id = object_id
            self.set_object_info(None)
            self._set_title("object @ " + thonny.memory.format_object_id(object_id))
            self.request_object_info()

    def _set_title(self, text):
        self.title_label.configure(text=text)

    def _handle_object_info_event(self, msg):
        if self.winfo_ismapped():
            if msg.info["id"] == self.object_id:
                if hasattr(msg, "not_found") and msg.not_found:
                    self.object_id = None
                    self.set_object_info(None)
                else:
                    self.set_object_info(msg.info)

    def _handle_progress_event(self, event):
        if self.object_id is not None:
            self.request_object_info()

    def request_object_info(self):
        # current width and height of the frame are required for
        # some content providers
        if self.active_page is not None:
            frame_width = self.active_page.winfo_width()
            frame_height = self.active_page.winfo_height()

            # in some cases measures are inaccurate
            if frame_width < 5 or frame_height < 5:
                frame_width = None
                frame_height = None
            # print("pa", frame_width, frame_height)
        else:
            frame_width = None
            frame_height = None

        get_runner().send_command(
            InlineCommand(
                "get_object_info",
                object_id=self.object_id,
                include_attributes=self.active_page == self.attributes_page,
                all_attributes=False,
                frame_width=frame_width,
                frame_height=frame_height,
            )
        )

    def set_object_info(self, object_info):
        self.object_info = object_info
        if object_info is None or "error" in object_info:
            if object_info is None:
                self._set_title("")
            else:
                self._set_title(object_info["error"])
            if self.current_content_inspector is not None:
                self.current_content_inspector.grid_remove()
                self.current_content_inspector = None
            self.attributes_page.clear()
        else:
            self._set_title(
                object_info["full_type_name"]
                + " @ "
                + thonny.memory.format_object_id(object_info["id"])
            )
            self.attributes_page.update_variables(object_info["attributes"])
            self.update_type_specific_info(object_info)

            # update layout
            # self._expose(None)
            # if not self.grid_frame.winfo_ismapped():
            #    self.grid_frame.grid()

        """
        if self.back_links == []:
            self.back_label.config(foreground="lightgray", cursor="arrow")
        else:
            self.back_label.config(foreground="blue", cursor="hand2")

        if self.forward_links == []:
            self.forward_label.config(foreground="lightgray", cursor="arrow")
        else:
            self.forward_label.config(foreground="blue", cursor="hand2")
        """

    def update_type_specific_info(self, object_info):
        content_inspector = None
        for insp in self.content_inspectors:
            if insp.applies_to(object_info):
                content_inspector = insp
                break

        # print("TYPSE", content_inspector)
        if content_inspector != self.current_content_inspector:
            if self.current_content_inspector is not None:
                self.current_content_inspector.grid_remove()  # TODO: or forget?
                self.current_content_inspector = None

            if content_inspector is not None:
                content_inspector.grid(row=0, column=0, sticky=tk.NSEW, padx=(0, 0))

            self.current_content_inspector = content_inspector

        if self.current_content_inspector is not None:
            self.current_content_inspector.set_object_info(object_info)

    def goto_type(self, event):
        if self.object_info is not None:
            get_workbench().event_generate("ObjectSelect", object_id=self.object_info["type_id"])


class ContentInspector:
    def __init__(self, master):
        pass

    def set_object_info(self, object_info):
        pass

    def get_tab_text(self):
        return "Data"

    def applies_to(self, object_info):
        return False


class FileHandleInspector(TextFrame, ContentInspector):
    def __init__(self, master):
        ContentInspector.__init__(self, master)
        TextFrame.__init__(self, master, read_only=True)
        self.cache = {}  # stores file contents for handle id-s
        self.config(borderwidth=1)
        self.text.configure(background="white")
        self.text.tag_configure("read", foreground="lightgray")

    def applies_to(self, object_info):
        return "file_content" in object_info or "file_error" in object_info

    def set_object_info(self, object_info):

        if "file_content" not in object_info:
            logging.exception("File error: " + object_info["file_error"])
            return

        assert "file_content" in object_info
        content = object_info["file_content"]
        line_count_sep = len(content.split("\n"))
        # line_count_term = len(content.splitlines())
        # char_count = len(content)
        self.text.configure(height=min(line_count_sep, 10))
        self.text.set_content(content)

        assert "file_tell" in object_info
        # f.tell() gives num of bytes read (minus some magic with linebreaks)

        file_bytes = content.encode(encoding=object_info["file_encoding"])
        bytes_read = file_bytes[0 : object_info["file_tell"]]
        read_content = bytes_read.decode(encoding=object_info["file_encoding"])
        read_char_count = len(read_content)
        # read_line_count_term = (len(content.splitlines())
        #                        - len(content[read_char_count:].splitlines()))

        pos_index = "1.0+" + str(read_char_count) + "c"
        self.text.tag_add("read", "1.0", pos_index)
        self.text.see(pos_index)

        # TODO: show this info somewhere
        """
        label.configure(text="Read %d/%d %s, %d/%d %s"
                        % (read_char_count,
                           char_count,
                           "symbol" if char_count == 1 else "symbols",
                           read_line_count_term,
                           line_count_term,
                           "line" if line_count_term == 1 else "lines"))
        """


class FunctionInspector(TextFrame, ContentInspector):
    def __init__(self, master):
        ContentInspector.__init__(self, master)
        TextFrame.__init__(self, master, read_only=True)
        self.text.configure(background="white")

    def applies_to(self, object_info):
        return "source" in object_info

    def get_tab_text(self):
        return "Code"

    def set_object_info(self, object_info):
        line_count = len(object_info["source"].split("\n"))
        self.text.configure(height=min(line_count, 15))
        self.text.set_content(object_info["source"])


class StringInspector(TextFrame, ContentInspector):
    def __init__(self, master):
        ContentInspector.__init__(self, master)
        TextFrame.__init__(self, master, read_only=True)
        # self.config(borderwidth=1)
        # self.text.configure(background="white")

    def applies_to(self, object_info):
        return object_info["type"] == repr(str)

    def set_object_info(self, object_info):
        # TODO: don't show too big string
        content = ast.literal_eval(object_info["repr"])
        line_count_sep = len(content.split("\n"))
        # line_count_term = len(content.splitlines())
        self.text.configure(height=min(line_count_sep, 10))
        self.text.set_content(content)
        """ TODO:
        label.configure(text="%d %s, %d %s"
                        % (len(content),
                           "symbol" if len(content) == 1 else "symbols",
                           line_count_term,
                           "line" if line_count_term == 1 else "lines"))
        """


class ReprInspector(TextFrame, ContentInspector):
    def __init__(self, master):
        ContentInspector.__init__(self, master)
        TextFrame.__init__(self, master, read_only=True)
        # self.config(borderwidth=1)
        # self.text.configure(background="white")

    def applies_to(self, object_info):
        return True

    def set_object_info(self, object_info):
        # TODO: don't show too big string
        content = object_info["repr"]
        self.text.set_content(content)
        """
        line_count_sep = len(content.split("\n"))
        line_count_term = len(content.splitlines())
        self.text.configure(height=min(line_count_sep, 10))
        label.configure(text="%d %s, %d %s"
                        % (len(content),
                           "symbol" if len(content) == 1 else "symbols",
                           line_count_term,
                           "line" if line_count_term == 1 else "lines"))
        """


class ElementsInspector(thonny.memory.MemoryFrame, ContentInspector):
    def __init__(self, master):
        ContentInspector.__init__(self, master)
        thonny.memory.MemoryFrame.__init__(self, master, ("index", "id", "value"))

        # self.vert_scrollbar.grid_remove()
        self.tree.column("index", width=40, anchor=tk.W, stretch=False)
        self.tree.column("id", width=750, anchor=tk.W, stretch=True)
        self.tree.column("value", width=750, anchor=tk.W, stretch=True)

        self.tree.heading("index", text="Index", anchor=tk.W)
        self.tree.heading("id", text="Value ID", anchor=tk.W)
        self.tree.heading("value", text="Value", anchor=tk.W)

        self.elements_have_indices = None
        self.update_memory_model()

        get_workbench().bind("ShowView", self.update_memory_model, True)
        get_workbench().bind("HideView", self.update_memory_model, True)

    def update_memory_model(self, event=None):
        self._update_columns()

    def _update_columns(self):
        if get_workbench().in_heap_mode():
            if self.elements_have_indices:
                self.tree.configure(displaycolumns=("index", "id"))
            else:
                self.tree.configure(displaycolumns=("id",))
        else:
            if self.elements_have_indices:
                self.tree.configure(displaycolumns=("index", "value"))
            else:
                self.tree.configure(displaycolumns=("value"))

    def applies_to(self, object_info):
        return "elements" in object_info

    def on_select(self, event):
        pass

    def on_double_click(self, event):
        self.show_selected_object_info()

    def set_object_info(self, object_info):
        assert "elements" in object_info

        self.elements_have_indices = object_info["type"] in (repr(tuple), repr(list))
        self._update_columns()

        self._clear_tree()
        index = 0
        # TODO: don't show too big number of elements
        for element in object_info["elements"]:
            node_id = self.tree.insert("", "end")
            if self.elements_have_indices:
                self.tree.set(node_id, "index", index)
            else:
                self.tree.set(node_id, "index", "")

            self.tree.set(node_id, "id", thonny.memory.format_object_id(element.id))
            self.tree.set(
                node_id, "value", shorten_repr(element.repr, thonny.memory.MAX_REPR_LENGTH_IN_GRID)
            )
            index += 1

        count = len(object_info["elements"])
        self.tree.config(height=min(count, 10))

        """ TODO:
        label.configure (
            text=("%d element" if count == 1 else "%d elements") % count
        )
        """


class DictInspector(thonny.memory.MemoryFrame, ContentInspector):
    def __init__(self, master):
        ContentInspector.__init__(self, master)
        thonny.memory.MemoryFrame.__init__(self, master, ("key_id", "id", "key", "value"))
        self.configure(border=1)
        # self.vert_scrollbar.grid_remove()
        self.tree.column("key_id", width=100, anchor=tk.W, stretch=False)
        self.tree.column("key", width=100, anchor=tk.W, stretch=False)
        self.tree.column("id", width=750, anchor=tk.W, stretch=True)
        self.tree.column("value", width=750, anchor=tk.W, stretch=True)

        self.tree.heading("key_id", text="Key ID", anchor=tk.W)
        self.tree.heading("key", text="Key", anchor=tk.W)
        self.tree.heading("id", text="Value ID", anchor=tk.W)
        self.tree.heading("value", text="Value", anchor=tk.W)

        self.update_memory_model()

    def update_memory_model(self, event=None):
        if get_workbench().in_heap_mode():
            self.tree.configure(displaycolumns=("key_id", "id"))
        else:
            self.tree.configure(displaycolumns=("key", "value"))

    def applies_to(self, object_info):
        return "entries" in object_info

    def on_select(self, event):
        pass

    def on_double_click(self, event):
        # NB! this selects value
        self.show_selected_object_info()

    def set_object_info(self, object_info):
        assert "entries" in object_info

        self._clear_tree()
        # TODO: don't show too big number of elements
        for key, value in object_info["entries"]:
            node_id = self.tree.insert("", "end")
            self.tree.set(node_id, "key_id", thonny.memory.format_object_id(key.id))
            self.tree.set(
                node_id, "key", shorten_repr(key.repr, thonny.memory.MAX_REPR_LENGTH_IN_GRID)
            )
            self.tree.set(node_id, "id", thonny.memory.format_object_id(value.id))
            self.tree.set(
                node_id, "value", shorten_repr(value.repr, thonny.memory.MAX_REPR_LENGTH_IN_GRID)
            )

        count = len(object_info["entries"])
        self.tree.config(height=min(count, 10))

        """ TODO:
        label.configure (
            text=("%d entry" if count == 1 else "%d entries") % count
        )
        """

        self.update_memory_model()


class ImageInspector(ContentInspector, tk.Frame):
    def __init__(self, master):
        tk.Frame.__init__(self, master)
        ContentInspector.__init__(self, master)
        self.label = tk.Label(self, anchor="nw")
        self.label.grid(row=0, column=0, sticky="nsew")
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

    def set_object_info(self, object_info):
        # print(object_info["image_data"])
        if isinstance(object_info["image_data"], bytes):
            data = base64.b64encode(object_info["image_data"])
        elif isinstance(object_info["image_data"], str):
            data = object_info["image_data"]
        else:
            self.label.configure(
                image=None, text="Unsupported image data (%s)" % type(object_info["image_data"])
            )
            return

        try:
            self.image = tk.PhotoImage(data=data)
            self.label.configure(image=self.image)
        except Exception as e:
            self.label.configure(image=None, text="Unsupported image data (%s)" % e)

    def applies_to(self, object_info):
        return "image_data" in object_info


class AttributesFrame(thonny.memory.VariablesFrame):
    def __init__(self, master):
        thonny.memory.VariablesFrame.__init__(self, master)
        self.configure(border=0)

    def on_select(self, event):
        pass

    def on_double_click(self, event):
        self.show_selected_object_info()


def load_plugin() -> None:
    get_workbench().add_view(ObjectInspector, _("Object inspector"), "se")



import ast
import logging
import io
import os
import platform
import queue
import re
import subprocess
import sys
import textwrap
import threading
import time
import tokenize
import traceback
import webbrowser
from queue import Queue
from textwrap import dedent
from time import sleep
from tkinter import ttk, messagebox
from thonny.ui_utils import askopenfilename, create_url_label
from typing import Optional

import jedi
import serial.tools.list_ports
from serial import SerialException

from thonny import common, get_runner, get_shell, get_workbench, running
from thonny.common import (
    BackendEvent,
    InlineResponse,
    MessageFromBackend,
    ToplevelCommand,
    ToplevelResponse,
    InterruptCommand,
    EOFCommand,
    CommandToBackend,
)
from thonny.config_ui import ConfigurationPage
from thonny.misc_utils import find_volumes_by_name, TimeHelper
from thonny.plugins.backend_config_page import BackendDetailsConfigPage
from thonny.running import BackendProxy, SubprocessProxy
from thonny.ui_utils import SubprocessDialog, create_string_var, show_dialog

DEFAULT_WEBREPL_URL = "ws://192.168.4.1:8266/"


class MicroPythonProxy(SubprocessProxy):
    def __init__(self, clean):
        self._port = get_workbench().get_option(self.backend_name + ".port")
        self._clean_start = clean

        if self._port == "auto":
            potential = self._detect_potential_ports()
            if len(potential) == 1:
                self._port = potential[0][0]
            else:
                self._port = None
                message = dedent(
                    """\
                    Couldn't find the device automatically. 
                    Check the connection (making sure the device is not in bootloader mode)
                    or choose "Tools → Options → Interpreter" to select the port manually."""
                )

                if len(potential) > 1:
                    _, descriptions = zip(*potential)
                    message += "\n\nLikely candidates are:\n * " + "\n * ".join(descriptions)

                self._show_error(message)

        super().__init__(clean, running.get_interpreter_for_subprocess())

        # Following is required for compatibility with older MP plugins (ESP)
        # TODO: remove it later
        self.micropython_upload_enabled = False

    def _get_launcher_with_args(self):
        import thonny.plugins.micropython.backend

        cmd = [
            thonny.plugins.micropython.backend.__file__,
            "--clean",
            str(self._clean_start),
            "--port",
            str(self._port),
            "--api_stubs_path",
            self._get_api_stubs_path(),
        ]

        if self._port == "webrepl":
            cmd.extend(
                [
                    "--url",
                    get_workbench().get_option(self.backend_name + ".webrepl_url"),
                    "--password",
                    get_workbench().get_option(self.backend_name + ".webrepl_password"),
                ]
            )

        return cmd

    def interrupt(self):
        self._send_msg(InterruptCommand())
        """
        # For some reason following kills the backend and KeyboardInterrupt can't be caught 
        if self._proc is not None and self._proc.poll() is None:
            if running_on_windows():
                try:
                    os.kill(self._proc.pid, signal.CTRL_BREAK_EVENT)  # @UndefinedVariable
                except Exception:
                    logging.exception("Could not interrupt backend process")
            else:
                self._proc.send_signal(signal.SIGINT)
        """

    def send_command(self, cmd: CommandToBackend) -> Optional[str]:
        if isinstance(cmd, EOFCommand):
            get_shell().restart()  # Runner doesn't notice restart

        return super().send_command(cmd)

    def _clear_environment(self):
        "TODO:"

    def _detect_potential_ports(self):
        all_ports = list_serial_ports()
        """
        for p in all_ports:
            print(p.description,
                  p.device,
                  None if p.vid is None else hex(p.vid),
                  None if p.pid is None else hex(p.pid),
                  )
        """
        return [
            (p.device, p.description)
            for p in all_ports
            if (p.vid, p.pid) in self.known_usb_vids_pids
            or p.description in self.known_port_descriptions
            or self.consider_unknown_devices
            and (
                ("USB" in p.description and "serial" in p.description.lower())
                or "UART" in p.description
                or "DAPLink" in p.description
                or "STLink" in p.description
            )
        ]

    @property
    def known_usb_vids_pids(self):
        """Return set of pairs of USB device VID, PID"""
        return set()

    @property
    def consider_unknown_devices(self):
        return True

    @property
    def known_port_descriptions(self):
        return set()

    def _get_api_stubs_path(self):
        import inspect

        return os.path.join(os.path.dirname(inspect.getfile(self.__class__)), "api_stubs")

    def supports_remote_files(self):
        return self._proc is not None

    def uses_local_filesystem(self):
        return False

    def ready_for_remote_file_operations(self):
        return self._proc is not None and get_runner().is_waiting_toplevel_command()

    def supports_remote_directories(self):
        return self._cwd is not None and self._cwd != ""

    def supports_trash(self):
        return False

    def is_connected(self):
        return self._proc is not None

    def _show_error(self, text):
        get_shell().print_error("\n" + text + "\n")

    def disconnect(self):
        self.destroy()

    def get_node_label(self):
        if "CircuitPython" in self._welcome_text:
            return _("CircuitPython device")
        elif "micro:bit" in self._welcome_text.lower():
            return "micro:bit"
        else:
            return _("MicroPython device")

    def get_exe_dirs(self):
        return []


class MicroPythonConfigPage(BackendDetailsConfigPage):
    backend_name = None  # Will be overwritten on Workbench.add_backend

    def __init__(self, master):
        super().__init__(master)
        intro_text = (
            _("Connect your device to the computer and select corresponding port below")
            + "\n"
            + "("
            + _('look for your device name, "USB Serial" or "UART"')
            + ").\n"
            + _("If you can't find it, you may need to install proper USB driver first.")
        )
        if self.allow_webrepl:
            intro_text = (
                ("Connecting via USB cable:")
                + "\n"
                + intro_text
                + "\n\n"
                + ("Connecting via WebREPL protocol:")
                + "\n"
                + (
                    "If your device supports WebREPL, first connect via serial, make sure WebREPL is enabled\n"
                    + "(import webrepl_setup), connect your computer and device to same network and select < WebREPL > below"
                )
            )

        intro_label = ttk.Label(self, text=intro_text)
        intro_label.grid(row=0, column=0, sticky="nw")

        driver_url = self._get_usb_driver_url()
        if driver_url:
            driver_url_label = create_url_label(self, driver_url)
            driver_url_label.grid(row=1, column=0, sticky="nw")

        port_label = ttk.Label(self, text="Port or WebREPL" if self.allow_webrepl else _("Port"))
        port_label.grid(row=3, column=0, sticky="nw", pady=(10, 0))

        self._ports_by_desc = {
            p.description
            if p.device in p.description
            else p.description + " (" + p.device + ")": p.device
            for p in list_serial_ports()
        }
        self._ports_by_desc["< " + _("Try to detect port automatically") + " >"] = "auto"

        self._WEBREPL_OPTION_DESC = "< WebREPL >"
        if self.allow_webrepl:
            self._ports_by_desc[self._WEBREPL_OPTION_DESC] = "webrepl"

        def port_order(p):
            _, name = p
            if name is None:
                return ""
            elif name.startswith("COM") and len(name) == 4:
                # Make one-digit COM ports go before COM10
                return name.replace("COM", "COM0")
            else:
                return name

        # order by port, auto first
        port_descriptions = [key for key, _ in sorted(self._ports_by_desc.items(), key=port_order)]

        self._port_desc_variable = create_string_var(
            self.get_stored_port_desc(), self._on_change_port
        )
        self._port_combo = ttk.Combobox(
            self,
            exportselection=False,
            textvariable=self._port_desc_variable,
            values=port_descriptions,
        )
        self._port_combo.state(["!disabled", "readonly"])

        self._port_combo.grid(row=4, column=0, sticky="new")
        self.columnconfigure(0, weight=1)

        self._webrepl_frame = None
        self._flashing_frame = None

        self._on_change_port()

    def _get_webrepl_frame(self):

        if self._webrepl_frame is not None:
            return self._webrepl_frame

        self._webrepl_frame = ttk.Frame(self)

        self._webrepl_url_var = create_string_var(
            get_workbench().get_option(self.backend_name + ".webrepl_url")
        )
        url_label = ttk.Label(self._webrepl_frame, text="URL (eg. %s)" % DEFAULT_WEBREPL_URL)
        url_label.grid(row=0, column=0, sticky="nw", pady=(10, 0))
        url_entry = ttk.Entry(self._webrepl_frame, textvariable=self._webrepl_url_var, width=24)
        url_entry.grid(row=1, column=0, sticky="nw")

        self._webrepl_password_var = create_string_var(
            get_workbench().get_option(self.backend_name + ".webrepl_password")
        )
        pw_label = ttk.Label(
            self._webrepl_frame, text="Password (the one specified with `import webrepl_setup`)"
        )
        pw_label.grid(row=2, column=0, sticky="nw", pady=(10, 0))
        pw_entry = ttk.Entry(self._webrepl_frame, textvariable=self._webrepl_password_var, width=9)
        pw_entry.grid(row=3, column=0, sticky="nw")

        return self._webrepl_frame

    def _get_flashing_frame(self):

        if self._flashing_frame is not None:
            return self._flashing_frame

        self._flashing_frame = ttk.Frame(self)
        self._flashing_label = ttk.Label(self._flashing_frame, text=_("Firmware"))
        self._flashing_label.grid(row=1, column=0, sticky="w", pady=(10, 0))

        button_caption = _("Open the dialog for installing or upgrading MicroPython on your device")
        self._flashing_button = ttk.Button(
            self._flashing_frame,
            text=button_caption,
            width=len(button_caption),
            command=self._open_flashing_dialog,
        )
        self._flashing_button.grid(row=2, column=0, sticky="we")

        self._flashing_frame.columnconfigure(0, weight=1)

        return self._flashing_frame

    def get_stored_port_desc(self):
        name = get_workbench().get_option(self.backend_name + ".port")
        for desc in self._ports_by_desc:
            if self._ports_by_desc[desc] == name:
                return desc

        return ""

    def get_selected_port_name(self):
        port_desc = self._port_desc_variable.get()
        return self._ports_by_desc[port_desc]

    def is_modified(self):
        return (
            self._port_desc_variable.modified  # pylint: disable=no-member
            or self.webrepl_selected()
            and self._webrepl_password_var.modified  # pylint: disable=no-member
            or self.webrepl_selected()
            and self._webrepl_url_var.modified
        )  # pylint: disable=no-member

    def webrepl_selected(self):
        return self.get_selected_port_name() == "webrepl"

    def should_restart(self):
        return self.is_modified()

    def apply(self):
        if not self.is_modified():
            return

        else:
            port_name = self.get_selected_port_name()
            get_workbench().set_option(self.backend_name + ".port", port_name)
            if self.webrepl_selected():
                get_workbench().set_option(
                    self.backend_name + ".webrepl_url", self._webrepl_url_var.get()
                )
                get_workbench().set_option(
                    self.backend_name + ".webrepl_password", self._webrepl_password_var.get()
                )

    def _on_change_port(self, *args):
        if self._port_desc_variable.get() == self._WEBREPL_OPTION_DESC:
            self._get_webrepl_frame().grid(row=6, column=0, sticky="nwe")
            if self._flashing_frame and self._flashing_frame.winfo_ismapped():
                self._flashing_frame.grid_forget()

        else:
            if self._has_flashing_dialog():
                self._get_flashing_frame().grid(row=6, column=0, sticky="nwe")

            if self._webrepl_frame and self._webrepl_frame.winfo_ismapped():
                self._webrepl_frame.grid_forget()

    def _get_usb_driver_url(self):
        return None

    def _has_flashing_dialog(self):
        return False

    def _open_flashing_dialog(self):
        raise NotImplementedError()

    @property
    def allow_webrepl(self):
        return False


class GenericMicroPythonProxy(MicroPythonProxy):
    @property
    def known_usb_vids_pids(self):
        """Return set of pairs of USB device (VID, PID)"""
        return {
            # Generic MicroPython Board, see http://pid.codes/org/MicroPython/
            (0x1209, 0xADDA)
        }


class GenericMicroPythonConfigPage(MicroPythonConfigPage):
    @property
    def allow_webrepl(self):
        return False


def list_serial_ports():
    # serial.tools.list_ports.comports() can be too slow
    # because os.path.islink can be too slow (https://github.com/pyserial/pyserial/pull/303)
    # Workarond: temporally patch os.path.islink
    try:
        old_islink = os.path.islink
        if platform.system() == "Windows":
            os.path.islink = lambda _: False
        return list(serial.tools.list_ports.comports())
    finally:
        os.path.islink = old_islink


def list_serial_ports_with_descriptions():
    def port_order(p):
        name = p.device
        if name is None:
            return ""
        elif name.startswith("COM") and len(name) == 4:
            # Make one-digit COM ports go before COM10
            return name.replace("COM", "COM0")
        else:
            return name

    sorted_ports = sorted(list_serial_ports(), key=port_order)

    return [
        (
            p.description if p.device in p.description else p.description + " (" + p.device + ")",
            p.device,
        )
        for p in sorted_ports
    ]


def add_micropython_backend(name, proxy_class, description, config_page):
    get_workbench().set_default(name + ".port", "auto")
    get_workbench().set_default(name + ".webrepl_url", DEFAULT_WEBREPL_URL)
    get_workbench().set_default(name + ".webrepl_password", "")
    get_workbench().add_backend(name, proxy_class, description, config_page)


def load_plugin():
    add_micropython_backend(
        "GenericMicroPython",
        GenericMicroPythonProxy,
        _("MicroPython (generic)"),
        GenericMicroPythonConfigPage,
    )

    def explain_deprecation():
        messagebox.showinfo(
            "Moved commands",
            dedent(
                """
            MicroPython commands have been moved or replaced:
            
            * "Select device"
                    * Moved into "Run" menu as "Select interpreter"
            * "Upload current script as main script"
            * "Upload current script as boot script"
            * "Upload current script with current name"
                    * use "File => Save copy..."
            * "Show device's main script"
            * "Show device's boot script"
                    * Double-click in Files view (Show => Files)
            * "Upload ____ firmware"
                    * Moved into interpreter config page or "Tools" menu
            * "Soft reboot"
                    * Moved into "Run" menu as "Send EOF / Soft reboot"
            " "Close serial connection"
                    * Moved into "Run" menu as "Disconnect"
        """
            ),
        )

    get_workbench().add_command(
        "devicedeprecation",
        "tempdevice",
        "Where are all the commands?",
        explain_deprecation,
        group=1,
    )



AP_IF = 1
AUTH_MAX = 6
AUTH_OPEN = 0
AUTH_WEP = 1
AUTH_WPA2_PSK = 3
AUTH_WPA_PSK = 2
AUTH_WPA_WPA2_PSK = 4


def LAN():
    pass


MODE_11B = 1
MODE_11G = 2
MODE_11N = 4
PHY_LAN8720 = 0
PHY_TLK110 = 1
STA_IF = 0


def WLAN():
    pass


def phy_mode():
    pass



import os.path
import sys
import platform
from typing import TYPE_CHECKING, cast, Optional
import traceback
import logging

SINGLE_INSTANCE_DEFAULT = True


def _compute_thonny_user_dir():
    if os.environ.get("THONNY_USER_DIR", ""):
        return os.path.expanduser(os.environ["THONNY_USER_DIR"])
    elif is_portable():
        if platform.system() == "Windows":
            root_dir = os.path.dirname(sys.executable)
        elif platform.system() == "Darwin":
            root_dir = os.path.join(
                os.path.dirname(sys.executable), "..", "..", "..", "..", "..", ".."
            )
        else:
            root_dir = os.path.join(os.path.dirname(sys.executable), "..")
        return os.path.normpath(os.path.abspath(os.path.join(root_dir, "user_data")))
    elif (
        hasattr(sys, "base_prefix")
        and sys.base_prefix != sys.prefix
        or hasattr(sys, "real_prefix")
        and getattr(sys, "real_prefix") != sys.prefix
    ):
        # we're in a virtualenv or venv
        return os.path.join(sys.prefix, ".thonny")
    elif platform.system() == "Windows":
        from thonny import misc_utils

        return os.path.join(misc_utils.get_roaming_appdata_dir(), "Thonny")
    elif platform.system() == "Darwin":
        return os.path.expanduser("~/Library/Thonny")
    else:
        # https://specifications.freedesktop.org/basedir-spec/latest/ar01s02.html
        data_home = os.environ.get(
            "XDG_CONFIG_HOME", os.path.expanduser(os.path.join("~", ".config"))
        )
        return os.path.join(data_home, "Thonny")


def is_portable():
    # it can be explicitly declared as portable or shared ...
    portable_marker_path = os.path.join(os.path.dirname(sys.executable), "portable_thonny.ini")
    shared_marker_path = os.path.join(os.path.dirname(sys.executable), "shared_thonny.ini")

    if os.path.exists(portable_marker_path) and not os.path.exists(shared_marker_path):
        return True
    elif not os.path.exists(portable_marker_path) and os.path.exists(shared_marker_path):
        return False

    # ... or it becomes implicitly portable if it's on a removable drive
    abs_location = os.path.abspath(__file__)
    if platform.system() == "Windows":
        drive = os.path.splitdrive(abs_location)[0]
        if drive.endswith(":"):
            from ctypes import windll

            return windll.kernel32.GetDriveTypeW(drive) == 2  # @UndefinedVariable
        else:
            return False
    elif platform.system() == "Darwin":
        # not exact heuristics
        return abs_location.startswith("/Volumes/")
    else:
        # not exact heuristics
        return abs_location.startswith("/media/") or abs_location.startswith("/mnt/")


def get_version():
    try:
        package_dir = os.path.dirname(sys.modules["thonny"].__file__)
        with open(os.path.join(package_dir, "VERSION"), encoding="ASCII") as fp:
            return fp.read().strip()
    except Exception:
        return "0.0.0"


THONNY_USER_DIR = _compute_thonny_user_dir()
CONFIGURATION_FILE = os.path.join(THONNY_USER_DIR, "configuration.ini")


def _get_ipc_file_path():
    from thonny import misc_utils
    import getpass

    if platform.system() == "Windows":
        base_dir = misc_utils.get_local_appdata_dir()
    else:
        base_dir = os.environ.get("XDG_RUNTIME_DIR")
        if not base_dir or not os.path.exists(base_dir):
            base_dir = os.environ.get("TMPDIR")

    if not base_dir or not os.path.exists(base_dir):
        base_dir = THONNY_USER_DIR

    ipc_dir = os.path.join(base_dir, "thonny-%s" % getpass.getuser())
    os.makedirs(ipc_dir, exist_ok=True)

    if not platform.system() == "Windows":
        os.chmod(ipc_dir, 0o700)

    return os.path.join(ipc_dir, "ipc.sock")


IPC_FILE = _get_ipc_file_path()


def _check_welcome():
    from thonny import misc_utils

    if not os.path.exists(CONFIGURATION_FILE) and not misc_utils.running_on_rpi():
        from thonny.first_run import FirstRunWindow
        from thonny.config import ConfigurationManager

        mgr = ConfigurationManager(CONFIGURATION_FILE)

        win = FirstRunWindow(mgr)
        win.mainloop()
        return win.ok
    else:
        return True


def launch():
    import gettext
    import runpy

    if sys.executable.endswith("thonny.exe"):
        # otherwise some library may try to run its subprocess with thonny.exe
        # NB! Must be pythonw.exe not python.exe, otherwise Runner thinks console
        # is already allocated.
        sys.executable = sys.executable[: -len("thonny.exe")] + "pythonw.exe"

    set_dpi_aware()

    try:
        runpy.run_module("thonny.customize", run_name="__main__")
    except ImportError:
        pass

    gettext.install("thonny", "locale")
    _prepare_thonny_user_dir()

    if not _check_welcome():
        return

    if _should_delegate():
        try:
            _delegate_to_existing_instance(sys.argv[1:])
            print("Delegated to an existing Thonny instance. Exiting now.")
            return 0
        except Exception:
            traceback.print_exc()

    # Did not or could not delegate

    try:
        from thonny import workbench

        bench = workbench.Workbench()
        try:
            bench.mainloop()
        except SystemExit:
            bench.destroy()
        return 0

    except SystemExit as e:
        from tkinter import messagebox

        messagebox.showerror("System exit", str(e))
        return -1

    except Exception:
        from logging import exception

        exception("Internal launch or mainloop error")
        from thonny import ui_utils

        dlg = ui_utils.LongTextDialog("Internal error", traceback.format_exc())
        ui_utils.show_dialog(dlg, get_workbench())
        return -1
    finally:
        runner = get_runner()
        if runner is not None:
            runner.destroy_backend()

    return 0


def _prepare_thonny_user_dir():
    if not os.path.exists(THONNY_USER_DIR):
        os.makedirs(THONNY_USER_DIR, mode=0o700, exist_ok=True)

        # user_dir_template is a post-installation means for providing
        # alternative default user environment in multi-user setups
        template_dir = os.path.join(os.path.dirname(__file__), "user_dir_template")

        if os.path.isdir(template_dir):
            import shutil

            def copy_contents(src_dir, dest_dir):
                # I want the copy to have current user permissions
                for name in os.listdir(src_dir):
                    src_item = os.path.join(src_dir, name)
                    dest_item = os.path.join(dest_dir, name)
                    if os.path.isdir(src_item):
                        os.makedirs(dest_item, mode=0o700)
                        copy_contents(src_item, dest_item)
                    else:
                        shutil.copyfile(src_item, dest_item)
                        os.chmod(dest_item, 0o600)

            copy_contents(template_dir, THONNY_USER_DIR)


def _should_delegate():
    if not os.path.exists(IPC_FILE):
        # no previous instance
        return

    from thonny.config import try_load_configuration

    configuration_manager = try_load_configuration(CONFIGURATION_FILE)
    configuration_manager.set_default("general.single_instance", SINGLE_INSTANCE_DEFAULT)
    return configuration_manager.get_option("general.single_instance")


def _delegate_to_existing_instance(args):
    from thonny import workbench
    import socket

    transformed_args = []
    for arg in args:
        if not arg.startswith("-"):
            arg = os.path.abspath(arg)

        transformed_args.append(arg)

    try:
        sock, secret = _create_client_socket()
    except Exception:
        # Maybe the lock is abandoned
        print("Trying to remove lock")
        os.remove(IPC_FILE)
        print("Successfully removed abandoned lock")
        raise

    data = repr((secret, transformed_args)).encode(encoding="utf_8")

    sock.settimeout(2.0)
    sock.sendall(data)
    sock.shutdown(socket.SHUT_WR)
    response = bytes([])
    while len(response) < len(workbench.SERVER_SUCCESS):
        new_data = sock.recv(2)
        if len(new_data) == 0:
            break
        else:
            response += new_data

    if response.decode("UTF-8") != workbench.SERVER_SUCCESS:
        raise RuntimeError("Unsuccessful delegation")


def _create_client_socket():
    import socket

    timeout = 2.0

    if platform.system() == "Windows":
        with open(IPC_FILE, "r") as fp:
            port = int(fp.readline().strip())
            secret = fp.readline().strip()

        # "localhost" can be much slower than "127.0.0.1"
        client_socket = socket.create_connection(("127.0.0.1", port), timeout=timeout)
    else:
        client_socket = socket.socket(socket.AF_UNIX)  # @UndefinedVariable
        client_socket.settimeout(timeout)
        client_socket.connect(IPC_FILE)
        secret = ""

    return client_socket, secret


def set_dpi_aware():
    # https://stackoverflow.com/questions/36134072/setprocessdpiaware-seems-not-to-work-under-windows-10
    # https://bugs.python.org/issue33656
    # https://msdn.microsoft.com/en-us/library/windows/desktop/dn280512(v=vs.85).aspx
    # https://github.com/python/cpython/blob/master/Lib/idlelib/pyshell.py
    if sys.platform == "win32":
        try:
            import ctypes

            PROCESS_SYSTEM_DPI_AWARE = 1
            ctypes.OleDLL("shcore").SetProcessDpiAwareness(PROCESS_SYSTEM_DPI_AWARE)
        except (ImportError, AttributeError, OSError):
            pass


if TYPE_CHECKING:
    # Following imports are required for MyPy
    # http://mypy.readthedocs.io/en/stable/common_issues.html#import-cycles
    import thonny.workbench
    from thonny.workbench import Workbench
    from thonny.running import Runner
    from thonny.shell import ShellView

_workbench = None  # type: Optional[Workbench]


def get_workbench() -> "Workbench":
    return cast("Workbench", _workbench)


_runner = None  # type: Optional[Runner]


def get_runner() -> "Runner":
    return cast("Runner", _runner)


def get_shell() -> "ShellView":
    return cast("ShellView", get_workbench().get_view("ShellView"))



class VfsFat:
    ""

    def chdir():
        pass

    def getcwd():
        pass

    def ilistdir():
        pass

    label = None

    def mkdir():
        pass

    def mkfs():
        pass

    def mount():
        pass

    def open():
        pass

    def remove():
        pass

    def rename():
        pass

    def rmdir():
        pass

    def stat():
        pass

    def statvfs():
        pass

    def umount():
        pass


def erase_filesystem():
    pass


def getmount():
    pass


def mount():
    pass


def remount():
    pass


def umount():
    pass



class VfsFat:
    ""

    def chdir():
        pass

    def getcwd():
        pass

    def ilistdir():
        pass

    def mkdir():
        pass

    def mkfs():
        pass

    def mount():
        pass

    def open():
        pass

    def remove():
        pass

    def rename():
        pass

    def rmdir():
        pass

    def stat():
        pass

    def statvfs():
        pass

    def umount():
        pass


def chdir():
    pass


def dupterm():
    pass


def dupterm_notify():
    pass


def getcwd():
    pass


def ilistdir():
    pass


def listdir():
    pass


def mkdir():
    pass


def mount():
    pass


def remove():
    pass


def rename():
    pass


def rmdir():
    pass


def stat():
    pass


def statvfs():
    pass


def umount():
    pass


def uname():
    pass


def urandom():
    pass
